#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2007-2008  Zsolt Foldvari
# Copyright (C) 2012       Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

__all__ = ["MaskedEntry", "ValidatableMaskedEntry"]

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext
import string
import sys

import logging

_LOG = logging.getLogger(".widgets.validatedmaskedentry")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import GObject
from gi.repository import GLib
from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import Pango

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.errors import MaskError, ValidationError, WindowActiveError
from .undoableentry import UndoableEntry
from gramps.gen.constfunc import is_quartz

# ============================================================================
#
# MaskedEntry and ValidatableMaskedEntry copied and merged from the Kiwi
# project's ValidatableProxyWidgetMixin, KiwiEntry and ProxyEntry.
#
# http://www.async.com.br/projects/kiwi
#
# ============================================================================

(DIRECTION_LEFT, DIRECTION_RIGHT) = (1, -1)

(INPUT_ASCII_LETTER, INPUT_ALPHA, INPUT_ALPHANUMERIC, INPUT_DIGIT) = list(range(4))

INPUT_FORMATS = {
    "0": INPUT_DIGIT,
    "L": INPUT_ASCII_LETTER,
    "A": INPUT_ALPHANUMERIC,
    "a": INPUT_ALPHANUMERIC,
    "&": INPUT_ALPHA,
}

# Todo list: Other useful Masks
#  9 - Digit, optional
#  ? - Ascii letter, optional
#  C - Alpha, optional

INPUT_CHAR_MAP = {
    INPUT_ASCII_LETTER: lambda text: text in string.ascii_letters,
    INPUT_ALPHA: str.isalpha,
    INPUT_ALPHANUMERIC: str.isalnum,
    INPUT_DIGIT: str.isdigit,
}

(COL_TEXT, COL_OBJECT) = list(range(2))


class MaskedEntry(UndoableEntry):
    """
    The MaskedEntry is an Entry subclass with additional features.

    Additional features:

      - Mask, force the input to meet certain requirements
      - IconEntry, allows you to have an icon inside the entry
      - convenience functions for completion

    .. note:: Gramps does not use the mask feature at the moment, so that code
              path is not tested
    """

    __gtype_name__ = "MaskedEntry"

    def __init__(self):
        self._block_changed = False
        UndoableEntry.__init__(self)

        # connect in UndoableEntry:
        # self.connect('insert-text', self._on_insert_text)
        # self.connect('delete-text', self._on_delete_text)
        self.connect_after("grab-focus", self._after_grab_focus)

        self.connect("changed", self._on_changed)

        self.connect("focus", self._on_focus)
        self.connect("focus-out-event", self._on_focus_out_event)
        self.connect("move-cursor", self._on_move_cursor)
        self.connect("button-press-event", self._on_button_press_event)
        self.connect("notify::cursor-position", self._on_notify_cursor_position)

        self._completion = None
        self._exact_completion = False
        self.hasicon = False
        ##        self._icon = IconEntry(self)

        # List of validators
        #  str -> static characters
        #  int -> dynamic, according to constants above
        self._mask_validators = []
        self._mask = None
        # Fields defined by mask
        # each item is a tuble, containing the begining and the end of the
        # field in the text
        self._mask_fields = []
        self._current_field = -1
        self._pos = 0
        self._selecting = False

        self._block_insert = False
        self._block_delete = False
        self.in_do_draw = False

    # Virtual methods, note do_size_alloc needs gtk 2.9 +
    ##    def do_size_allocate(self, allocation):
    ##        Gtk.Entry.do_size_allocate(self, allocation)
    ##
    ##        if self.get_realized():
    ##            self._icon.resize_windows()

    ##    def do_draw(self, cairo_t):
    ##        Gtk.Entry.do_draw(self, cairo_t)
    ##
    ##        if Gtk.cairo_should_draw_window(cairo_t, self.get_window()):
    ##            self._icon.draw_pixbuf()

    ##    def do_realize(self):
    ##        Gtk.Entry.do_realize(self)
    ##        self._icon.construct()

    ##    def do_unrealize(self):
    ##        self._icon.deconstruct()
    ##        Gtk.Entry.do_unrealize(self)

    # Mask & Fields

    def set_mask(self, mask):
        """
        Set the mask of the Entry.

        Supported format characters are:

          - '0' digit
          - 'L' ascii letter (a-z and A-Z)
          - '&' alphabet, honors the locale
          - 'a' alphanumeric, honors the locale
          - 'A' alphanumeric, honors the locale

        This is similar to MaskedTextBox:
        http://msdn2.microsoft.com/en-us/library/system.windows.forms.maskedtextbox.mask(VS.80).aspx

        Example mask for a ISO-8601 date
        >>> entry.set_mask('0000-00-00')

        :param mask: the mask to set
        """
        if not mask:
            # self.override_font(Pango.FontDescription("sans"))
            self._mask = mask
            return

        # First, reset
        self._mask_validators = []
        self._mask_fields = []
        self._current_field = -1

        mask = str(mask)
        input_length = len(mask)
        lenght = 0
        pos = 0
        field_begin = 0
        field_end = 0
        while True:
            if pos >= input_length:
                break
            if mask[pos] in INPUT_FORMATS:
                self._mask_validators += [INPUT_FORMATS[mask[pos]]]
                field_end += 1
            else:
                self._mask_validators.append(mask[pos])
                if field_begin != field_end:
                    self._mask_fields.append((field_begin, field_end))
                field_end += 1
                field_begin = field_end
            pos += 1

        self._mask_fields.append((field_begin, field_end))
        # The set_mask function doesn't seem to be used, except for the test
        # so removing the monospace doesn't change visible functionality
        # self.override_font(Pango.FontDescription("monospace"))

        self._really_delete_text(0, -1)
        self._insert_mask(0, input_length)
        self._mask = mask

    def get_mask(self):
        """
        :returns: the mask
        """
        return self._mask

    def get_field_text(self, field):
        if not self._mask:
            raise MaskError("a mask must be set before calling get_field_text")
        # assert self._mask
        text = self.get_text()
        start, end = self._mask_fields[field]
        return text[start:end].strip()

    def get_fields(self):
        """
        Get the fields assosiated with the entry.
        A field is dynamic content separated by static.
        For example, the format string 000-000 has two fields
        separated by a dash.
        if a field is empty it'll return an empty string
        otherwise it'll include the content

        :returns: fields
        :rtype: list of strings
        """
        if not self._mask:
            raise MaskError("a mask must be set before calling get_fields")
        # assert self._mask

        fields = []

        text = str(self.get_text())
        for start, end in self._mask_fields:
            fields.append(text[start:end].strip())

        return fields

    def get_empty_mask(self, start=None, end=None):
        """
        Get the empty mask between start and end

        :param start:
        :param end:
        :returns: mask
        :rtype: string
        """

        if start is None:
            start = 0
        if end is None:
            end = len(self._mask_validators)

        s = ""
        for validator in self._mask_validators[start:end]:
            if isinstance(validator, int):
                s += " "
            elif isinstance(validator, str):
                s += validator
            else:
                raise AssertionError
        return s

    def get_field_pos(self, field):
        """
        Get the position at the specified field.
        """
        if field >= len(self._mask_fields):
            return None

        start, end = self._mask_fields[field]

        return start

    def _get_field_ideal_pos(self, field):
        start, end = self._mask_fields[field]
        text = self.get_field_text(field)
        pos = start + len(text)
        return pos

    def get_field(self):
        if self._current_field >= 0:
            return self._current_field
        else:
            return None

    def set_field(self, field, select=False):
        if field >= len(self._mask_fields):
            return

        pos = self._get_field_ideal_pos(field)
        self.set_position(pos)

        if select:
            field_text = self.get_field_text(field)
            start, end = self._mask_fields[field]
            self.select_region(start, pos)

        self._current_field = field

    def get_field_length(self, field):
        if 0 <= field < len(self._mask_fields):
            start, end = self._mask_fields[field]
            return end - start

    def _shift_text(self, start, end, direction=DIRECTION_LEFT, positions=1):
        """
        Shift the text, to the right or left, n positions. Note that this
        does not change the entry text. It returns the shifted text.

        :param start:
        :param end:
        :param direction: DIRECTION_LEFT or DIRECTION_RIGHT
        :param positions: the number of positions to shift.
        :returns: returns the text between start and end, shifted to the
                  direction provided.
        """
        text = self.get_text()
        new_text = ""
        validators = self._mask_validators

        if direction == DIRECTION_LEFT:
            i = start
        else:
            i = end - 1

        # When shifting a text, we wanna keep the static chars where they
        # are, and move the non-static chars to the right position.
        while start <= i < end:
            if isinstance(validators[i], int):
                # Non-static char shoud be here. Get the next one (depending
                # on the direction, and the number of positions to skip.)
                #
                # When shifting left, the next char will be on the right,
                # so, it will be appended, to the new text.
                # Otherwise, when shifting right, the char will be
                # prepended.
                next_pos = self._get_next_non_static_char_pos(
                    i, direction, positions - 1
                )

                # If its outside the bounds of the region, ignore it.
                if not start <= next_pos <= end:
                    next_pos = None

                if next_pos is not None:
                    if direction == DIRECTION_LEFT:
                        new_text = new_text + text[next_pos]
                    else:
                        new_text = text[next_pos] + new_text
                else:
                    if direction == DIRECTION_LEFT:
                        new_text = new_text + " "
                    else:
                        new_text = " " + new_text

            else:
                # Keep the static char where it is.
                if direction == DIRECTION_LEFT:
                    new_text = new_text + text[i]
                else:
                    new_text = text[i] + new_text
            i += direction

        return new_text

    def _get_next_non_static_char_pos(self, pos, direction=DIRECTION_LEFT, skip=0):
        """
        Get next non-static char position, skiping some chars, if necessary.

        :param skip: skip first n chars
        :param direction: direction of the search.
        """
        text = self.get_text()
        validators = self._mask_validators
        i = pos + direction + skip
        while 0 <= i < len(text):
            if isinstance(validators[i], int):
                return i
            i += direction

        return None

    def _get_field_at_pos(self, pos, dir=None):
        """
        Return the field index at position pos.
        """
        for p in self._mask_fields:
            if p[0] <= pos <= p[1]:
                return self._mask_fields.index(p)

        return None

    def set_exact_completion(self, value):
        """
        Enable exact entry completion.
        Exact means it needs to start with the value typed
        and the case needs to be correct.

        :param value: enable exact completion
        :type value: boolean
        """

        self._exact_completion = value
        if value:
            match_func = self._completion_exact_match_func
        else:
            match_func = self._completion_normal_match_func
        completion = self._get_completion()
        completion.set_match_func(match_func, None)

    def is_empty(self):
        text = self.get_text()
        if self._mask:
            empty = self.get_empty_mask()
        else:
            empty = ""

        return text == empty

    # Private

    def _really_delete_text(self, start, end):
        # A variant of delete_text() that never is blocked by us
        self._block_delete = True
        self.delete_text(start, end)
        self._block_delete = False

    def _really_insert_text(self, text, position):
        # A variant of insert_text() that never is blocked by us
        self._block_insert = True
        self.insert_text(text, position)
        self._block_insert = False

    def _insert_mask(self, start, end):
        text = self.get_empty_mask(start, end)
        self._really_insert_text(text, position=start)

    def _confirms_to_mask(self, position, text):
        validators = self._mask_validators
        if position < 0 or position >= len(validators):
            return False

        validator = validators[position]
        if isinstance(validator, int):
            if not INPUT_CHAR_MAP[validator](text):
                return False
        if isinstance(validator, str):
            if validator == text:
                return True
            return False

        return True

    def _get_completion(self):
        # Check so we have completion enabled, not this does not
        # depend on the property, the user can manually override it,
        # as long as there is a completion object set
        completion = self.get_completion()
        if completion:
            return completion

        completion = Gtk.EntryCompletion()
        self.set_completion(completion)
        return completion

    def get_completion(self):
        return self._completion

    def set_completion(self, completion):
        Gtk.Entry.set_completion(self, completion)
        # FIXME objects not supported yet, should it be at all?
        # completion.set_model(Gtk.ListStore(str, object))
        completion.set_model(Gtk.ListStore(GObject.TYPE_STRING))
        completion.set_text_column(0)
        # completion.connect("match-selected",
        # self._on_completion__match_selected)

        self._completion = Gtk.Entry.get_completion(self)
        self.set_exact_completion(self._exact_completion)
        return

    def set_completion_mode(self, popup=None, inline=None):
        """
        Set the way how completion is presented.

        :param popup: enable completion in popup window
        :type popup: boolean
        :param inline: enable inline completion
        :type inline: boolean
        """
        completion = self._get_completion()
        if popup is not None:
            completion.set_popup_completion(popup)
        if inline is not None:
            completion.set_inline_completion(inline)

    def _completion_exact_match_func(self, completion, key, iter):
        model = completion.get_model()
        if not len(model):
            return

        content = model[iter][COL_TEXT]
        return content.startswith(self.get_text())

    def _completion_normal_match_func(self, completion, key, iter, data=None):
        model = completion.get_model()
        if not len(model):
            return

        content = model[iter][COL_TEXT].lower()
        return key.lower() in content

    def _on_completion__match_selected(self, completion, model, iter, data=None):
        if not len(model):
            return

        # this updates current_object and triggers content-changed
        self.set_text(model[iter][COL_TEXT])
        self.set_position(-1)
        # FIXME: Enable this at some point
        # self.activate()

    def _appers_later(self, char, start):
        """
        Check if a char appers later on the mask. If it does, return
        the field it appers at. returns False otherwise.
        """
        validators = self._mask_validators
        i = start
        while i < len(validators):
            if self._mask_validators[i] == char:
                field = self._get_field_at_pos(i)
                if field is None:
                    return False

                return field

            i += 1

        return False

    def _can_insert_at_pos(self, new, pos):
        """
        Check if a chararcter can be inserted at some position

        :param new: The char that wants to be inserted.
        :param pos: The position where it wants to be inserted.
        :returns: Returns None if it can be inserted. If it cannot be,
                  return the next position where it can be successfuly
                  inserted.
        """
        validators = self._mask_validators

        # Do not let insert if the field is full
        field = self._get_field_at_pos(pos)
        if field is not None:
            text = self.get_field_text(field)
            length = self.get_field_length(field)
            if len(text) == length:
                Gdk.beep()
                return pos

        # If the char confirms to the mask, but is a static char, return the
        # position after that static char.
        if self._confirms_to_mask(pos, new) and not isinstance(validators[pos], int):
            return pos + 1

        # If does not confirms to mask:
        #  - Check if the char the user just tried to enter appers later.
        #  - If it does, Jump to the start of the field after that
        if not self._confirms_to_mask(pos, new):
            field = self._appers_later(new, pos)
            if field is not False:
                pos = self.get_field_pos(field + 1)
                if pos is not None:
                    GLib.idle_add(self.set_position, pos)
            return pos

        return None

    #   When inserting new text, supose, the entry, at some time is like this,
    #   ahd the user presses '0', for instance:
    #   --------------------------------
    #   | ( 1 2 )   3 4 5   - 6 7 8 9  |
    #   --------------------------------
    #              ^ ^     ^
    #              S P     E
    #
    #   S - start of the field (start)
    #   E - end of the field (end)
    #   P - pos - where the new text is being inserted. (pos)
    #
    #   So, the new text will be:
    #
    #     the old text, from 0 until P
    #   + the new text
    #   + the old text, from P until the end of the field, shifted to the
    #     right
    #   + the old text, from the end of the field, to the end of the text.
    #
    #   After inserting, the text will be this:
    #   --------------------------------
    #   | ( 1 2 )   3 0 4 5 - 6 7 8 9  |
    #   --------------------------------
    #              ^   ^   ^
    #              S   P   E
    #

    def _insert_at_pos(self, text, new, pos):
        """
        Inserts the character at the give position in text. Note that the
        insertion won't be applied to the entry, but to the text provided.

        :param text: Text that it will be inserted into.
        :param new: New text to insert.
        :param pos: Positon to insert at
        :returns: Returns a tuple, with the position after the insetion and the
                  new text.
        """
        field = self._get_field_at_pos(pos)
        length = len(new)
        new_pos = pos
        start, end = self._mask_fields[field]

        # Shift Right
        new_text = (
            text[:pos]
            + new
            + self._shift_text(pos, end, DIRECTION_RIGHT)[1:]
            + text[end:]
        )

        # Overwrite Right
        #        new_text = (text[:pos] + new +
        #                    text[pos+length:end]+
        #                    text[end:])
        new_pos = pos + 1
        GLib.idle_add(self.set_position, new_pos)

        # If the field is full, jump to the next field
        if len(self.get_field_text(field)) == self.get_field_length(field) - 1:
            GLib.idle_add(self.set_field, field + 1, True)
            self.set_field(field + 1)

        return new_pos, new_text

    # Callbacks
    def _on_insert_text(self, editable, new, length, position):
        if self._block_insert:
            return
        if not self._mask:
            UndoableEntry._on_insert_text(self, editable, new, length, position)
            return
        new = str(new)
        pos = self.get_position()

        self.stop_emission("insert-text")

        text = self.get_text()
        # Insert one char at a time
        for c in new:
            _pos = self._can_insert_at_pos(c, pos)
            if _pos is None:
                pos, text = self._insert_at_pos(text, c, pos)
            else:
                pos = _pos

        # Change the text with the new text.
        self._block_changed = True
        self._really_delete_text(0, -1)
        ### mask not used in Gramps, following should work though
        ##UndoableEntry._on_delete_text(self, editable, 0, -1)
        self._block_changed = False

        self._really_insert_text(text, 0)
        ### mask not used in Gramps, following should work though
        ##UndoableEntry._on_insert_text(self, editable, text, len(text),0)

    #   When deleting some text, supose, the entry, at some time is like this:
    #   --------------------------------
    #   | ( 1 2 )   3 4 5 6 - 7 8 9 0  |
    #   --------------------------------
    #              ^ ^ ^   ^
    #              S s e   E
    #
    #   S - start of the field (_start)
    #   E - end of the field (_end)
    #   s - start of the text being deleted (start)
    #   e - end of the text being deleted (end)
    #
    #   end - start -> the number of characters being deleted.
    #
    #   So, the new text will be:
    #
    #     the old text, from 0 until the start of the text being deleted.
    #   + the old text, from the start of where the text is being deleted, until
    #     the end of the field, shifted to the left, end-start positions
    #   + the old text, from the end of the field, to the end of the text.
    #
    #   So, after the text is deleted, the entry will look like this:
    #
    #   --------------------------------
    #   | ( 1 2 )   3 5 6   - 7 8 9 0  |
    #   --------------------------------
    #                ^
    #                P
    #
    #   P = the position of the cursor after the deletion, witch is equal to
    #   start (s at the previous ilustration)

    def _on_delete_text(self, editable, start, end):
        if self._block_delete:
            return
        if not self._mask:
            UndoableEntry._on_delete_text(self, editable, start, end)
            return

        self.stop_emission("delete-text")

        pos = self.get_position()
        # Trying to delete an static char. Delete the char before that
        if (
            0 < start < len(self._mask_validators)
            and not isinstance(self._mask_validators[start], int)
            and pos != start
        ):
            self._on_delete_text(editable, start - 1, start)
            ### mask not used in Gramps, following should work though
            ##UndoableEntry._on_delete_text(self, editable, start-1, start)
            return

        field = self._get_field_at_pos(end - 1)
        # Outside a field. Cannot delete.
        if field is None:
            self.set_position(end - 1)
            return
        _start, _end = self._mask_fields[field]

        # Deleting from outside the bounds of the field.
        if start < _start or end > _end:
            _start, _end = start, end

        # Change the text
        text = self.get_text()

        # Shift Left
        new_text = (
            text[:start]
            + self._shift_text(start, _end, DIRECTION_LEFT, end - start)
            + text[_end:]
        )

        # Overwrite Left
        #        empty_mask = self.get_empty_mask()
        #        new_text = (text[:_start] +
        #                    text[_start:start] +
        #                    empty_mask[start:start+(end-start)] +
        #                    text[start+(end-start):_end] +
        #                    text[_end:])

        new_pos = start

        self._block_changed = True
        self._really_delete_text(0, -1)
        ### mask not used in Gramps, following should work though
        ##UndoableEntry._on_delete_text(self, editable, 0, -1)
        self._block_changed = False
        self._really_insert_text(new_text, 0)
        ### mask not used in Gramps, following should work though
        ##UndoableEntry._on_insert_text(self, editable, text, len(text),0)

        # Position the cursor on the right place.
        self.set_position(new_pos)

        if self.is_empty():
            pos = self.get_field_pos(0)
            self.set_position(pos)

    def _after_grab_focus(self, widget):
        # The text is selectet in grab-focus, so this needs to be done after
        # that:
        if self.is_empty():
            if self._mask:
                self.set_field(0)
            else:
                self.set_position(0)

    def _on_focus(self, widget, direction):
        if not self._mask:
            return

        if direction == Gtk.DIR_TAB_FORWARD or direction == Gtk.DIR_DOWN:
            inc = 1
        if direction == Gtk.DIR_TAB_BACKWARD or direction == Gtk.DIR_UP:
            inc = -1

        field = self._current_field

        field += inc
        # Leaving the entry
        if field == len(self._mask_fields) or field == -1:
            self.select_region(0, 0)
            self._current_field = -1
            return False

        if field < 0:
            field = len(self._mask_fields) - 1

        # grab_focus changes the selection, so we need to grab_focus before
        # making the selection.
        self.grab_focus()
        self.set_field(field, select=True)

        return True

    def _on_notify_cursor_position(self, widget, pspec):
        if not self._mask:
            return

        if not self.is_focus():
            return

        if self._selecting:
            return

        pos = self.get_position()
        field = self._get_field_at_pos(pos)

        if pos == 0:
            self.set_position(self.get_field_pos(0))
            return

        text = self.get_text()
        field = self._get_field_at_pos(pos)

        # Humm, the pos is not inside any field. Get the next pos inside
        # some field, depending on the direction that the cursor is
        # moving
        diff = pos - self._pos
        _field = field
        while _field is None and (len(text) > pos > 0) and diff:
            pos += diff
            _field = self._get_field_at_pos(pos)
            self._pos = pos

        if field is None:
            self.set_position(self._pos)
        else:
            self._current_field = field
            self._pos = pos

    def _on_changed(self, widget):
        if self._block_changed:
            self.stop_emission("changed")

    def _on_focus_out_event(self, widget, event):
        if not self._mask:
            return

        self._current_field = -1

    def _on_move_cursor(self, entry, step, count, extend_selection):
        self._selecting = extend_selection

    def _on_button_press_event(self, entry, event):
        if event.type == Gdk.EventType.BUTTON_PRESS and event.button == 1:
            self._selecting = True
        elif event.type == Gdk.EventType.BUTTON_RELEASE and event.button == 1:
            self._selecting = True

    # IconEntry

    def set_tooltip(self, text):
        self.set_icon_tooltip_text(Gtk.EntryIconPosition.SECONDARY, text)

    def set_pixbuf(self, pixbuf):
        self.set_icon_from_pixbuf(Gtk.EntryIconPosition.SECONDARY, pixbuf)

    def set_stock(self, icon_name):
        self.set_icon_from_icon_name(Gtk.EntryIconPosition.SECONDARY, icon_name)

    # Gtk.EntryCompletion convenience function

    def prefill(self, itemdata, sort=False):
        if not isinstance(itemdata, (list, tuple)):
            raise TypeError(
                "'data' parameter must be a list or tuple of item "
                "descriptions, found %s"
            ) % type(itemdata)

        completion = self._get_completion()
        model = completion.get_model()

        if len(itemdata) == 0:
            model.clear()
            return

        values = {}
        if sort:
            itemdata.sort()

        for item in itemdata:
            if item in values:
                raise KeyError(
                    "Tried to insert duplicate value " "%r into the entry" % item
                )
            else:
                values[item] = None

            model.append((item,))


# number = (int, float, long)

VALIDATION_ICON_WIDTH = 16
MANDATORY_ICON = "dialog-information"
ERROR_ICON = "process-stop"
DELAY_TIME = 2500
READWRITE = (
    GObject.PARAM_READWRITE
    if GLib.check_version(2, 42, 0)
    else GObject.ParamFlags.READWRITE
)


class ValidatableMaskedEntry(MaskedEntry):
    """
    It extends the MaskedEntry with validation feature.

    Merged from Kiwi's ValidatableProxyWidgetMixin and ProxyEntry.
    To provide custom validation connect to the 'validate' signal
    of the instance.
    """

    __gtype_name__ = "ValidatableMaskedEntry"

    __gsignals__ = {
        "content-changed": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "validation-changed": (
            GObject.SignalFlags.RUN_FIRST,
            None,
            (GObject.TYPE_BOOLEAN,),
        ),
        "validate": (
            GObject.SignalFlags.RUN_LAST,
            GObject.TYPE_PYOBJECT,
            (GObject.TYPE_PYOBJECT,),
        ),
        "changed": "override",
    }

    __gproperties__ = {
        "data-type": (
            GObject.TYPE_PYOBJECT,
            "Data Type of the widget",
            "Type object",
            READWRITE,
        ),
        "mandatory": (GObject.TYPE_BOOLEAN, "Mandatory", "Mandatory", False, READWRITE),
    }

    # FIXME put the data type support back
    # allowed_data_types = (basestring, datetime.date, datetime.time,
    # datetime.datetime, object) + number

    def __init__(self, data_type=None, error_icon=ERROR_ICON):
        self.data_type = None
        self.mandatory = False
        self.error_icon = error_icon
        self._block_changed = False

        MaskedEntry.__init__(self)

        self._valid = True
        self._def_error_msg = None
        self.__delay_tag = None

        # FIXME put data type support back
        # self.set_property('data-type', data_type)

    # Virtual methods
    def do_changed(self):
        block_changed = getattr(self, "_block_changed", True)
        if block_changed:
            self.stop_emission_by_name("changed")
            return
        self.emit("content-changed")
        self.validate()

    def do_get_property(self, prop):
        """Return the gproperty's value."""

        if prop.name == "data-type":
            return self.data_type
        elif prop.name == "mandatory":
            return self.mandatory
        else:
            raise AttributeError("unknown property %s" % prop.name)

    def do_set_property(self, prop, value):
        """Set the property of writable properties."""

        if prop.name == "data-type":
            if value is None:
                self.data_type = value
                return

            # FIXME put the data type support back
            # if not issubclass(value, self.allowed_data_types):
            # raise TypeError(
            # "%s only accept %s types, not %r"
            # % (self,
            #' or '.join([t.__name__ for t in self.allowed_data_types]),
            # value))
            self.data_type = value
        elif prop.name == "mandatory":
            self.mandatory = value
        else:
            raise AttributeError("unknown or read only property %s" % prop.name)

    # Public API

    def set_default_error_msg(self, text):
        """
        Set default message for validation error.

        Default error message for an instance is useful when completion is
        used, because this case custom validation is not called.

        :param text: can contain one and only one '%s', where the actual value
                     of the Entry will be inserted.
        :type text: str
        """
        if not isinstance(text, str):
            raise TypeError("text must be a string")

        self._def_error_msg = text

    def is_valid(self):
        """
        :returns: True if the widget is in validated state
        """
        return self._valid

    def validate(self, force=False):
        """
        Checks if the data is valid.
        Validates data-type and custom validation.

        :param force: if True, force validation
        :returns: validated data or ValueUnset if it failed
        """

        # If we're not visible or sensitive return a blank value, except
        # when forcing the validation
        if not force and (
            not self.get_property("visible") or not self.get_property("sensitive")
        ):
            return None

        try:
            text = self.get_text()
            ##_LOG.debug('Read %r for %s' %  (data, self.model_attribute))

            # check if we should draw the mandatory icon
            # this need to be done before any data conversion because we
            # we don't want to end drawing two icons
            if self.mandatory and self.is_empty():
                self.set_blank()
                return None
            else:
                if self._completion:
                    for row in self.get_completion().get_model():
                        if row[COL_TEXT] == text:
                            break
                    else:
                        if text:
                            raise ValidationError()
                else:
                    if not self.is_empty():
                        # this signal calls the custom validation method
                        # of the instance and gets the exception (if any).
                        error = self.emit("validate", text)
                        if error:
                            raise error

            self.set_valid()
            return text
        except ValidationError as e:
            self.set_invalid(str(e))
            return None

    def set_valid(self):
        """
        Change the validation state to valid, which will remove icons and
        reset the background color
        """
        ##_LOG.debug('Setting state for %s to VALID' % self.model_attribute)
        if self.is_valid():
            return

        self._set_valid_state(True)

        if self.__delay_tag is not None:
            GLib.source_remove(self.__delay_tag)
            self.__delay_tag = None
        self.set_stock(None)
        context = self.get_style_context()
        context.remove_class("error")

    def set_invalid(self, text=None):
        """
        Change the validation state to invalid.

        :param text: text of tooltip of None
        """
        ##_LOG.debug('Setting state for %s to INVALID' % self.model_attribute)
        if not self.is_valid():
            return

        self._set_valid_state(False)

        generic_text = (
            _("'%s' is not a valid value " "for this field") % self.get_text()
        )

        # If there is no error text, let's try with the default or
        # fall back to a generic one
        if not text:
            text = self._def_error_msg
        if not text:
            text = generic_text

        try:
            text.index("%s")
            text = text % self.get_text()
        except TypeError:
            # if text contains '%s' more than once
            _LOG.error(
                'There must be only one instance of "%s"' " in validation error message"
            )
            # fall back to a generic one so the error icon still have a tooltip
            text = generic_text
        except ValueError:
            # if text does not contain '%s'
            pass

        self.set_tooltip(text)

        context = self.get_style_context()
        self.__delay_tag = GLib.timeout_add(DELAY_TIME, self.__delay_finished)
        context.add_class("error")

    def set_blank(self):
        """
        Change the validation state to blank state, this only applies
        for mandatory widgets, draw an icon and set a tooltip
        """

        ##_LOG.debug('Setting state for %s to BLANK' % self.model_attribute)

        if self.mandatory:
            self.set_stock(MANDATORY_ICON)
            self.set_tooltip(_("This field is mandatory"))
            valid = False
        else:
            valid = True

        self._set_valid_state(valid)

    def set_text(self, text):
        """
        Set the text of the entry

        :param text:
        """

        # If content isn't empty set_text emitts changed twice.
        # Protect content-changed from being updated and issue
        # a manual emission afterwards
        pos = self.get_position()
        self._block_changed = True
        MaskedEntry.set_text(self, text)
        self._block_changed = False
        self.emit("content-changed")

        if pos < len(text):
            self.set_position(pos)
        else:
            self.set_position(-1)

    # Private

    def _set_valid_state(self, state):
        """Updates the validation state and emits a signal if it changed"""

        if self._valid == state:
            return

        self.emit("validation-changed", state)
        self._valid = state

    def __delay_finished(self):
        """Set error icon after delay has finished."""
        self.__delay_tag = None
        self.set_stock(self.error_icon)
        return False


def main(args):
    from gramps.gen.datehandler import parser

    def on_validate(widget, text):
        myDate = parser.parse(text)
        if not myDate.is_regular():
            # used on AgeOnDateGramplet
            return ValidationError(_("'%s' is not a valid date value"))

    win = Gtk.Window()
    win.set_title("ValidatableMaskedEntry test window")
    win.set_position(Gtk.WindowPosition.CENTER)
    # Set the mnemonic modifier on Macs to alt-ctrl so that it
    # doesn't interfere with the extended keyboard, see
    # https://gramps-project.org/bugs/view.php?id=6943
    if is_quartz():
        win.set_mnemonic_modifier(
            Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.MOD1_MASK
        )

    def cb(window, event):
        Gtk.main_quit()

    win.connect("delete-event", cb)

    vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
    win.add(vbox)

    label = Gtk.Label(label="Pre-filled entry validated against the given list:")
    vbox.pack_start(label, True, True, 0)

    widget1 = ValidatableMaskedEntry(str)
    widget1.set_completion_mode(inline=True, popup=False)
    widget1.set_default_error_msg("'%s' is not a default Event")
    # widget1.set_default_error_msg(widget1)
    widget1.prefill(("Birth", "Death", "Conseption"))
    # widget1.set_exact_completion(True)
    vbox.pack_start(widget1, True, False, 0)

    label = Gtk.Label(label="Mandatory masked entry validated against user function:")
    vbox.pack_start(label, True, True, 0)

    # widget2 = ValidatableMaskedEntry(str, "#e0e0e0", error_icon=None)
    widget2 = ValidatableMaskedEntry()
    widget2.set_mask("00/00/0000")
    widget2.connect("validate", on_validate)
    widget2.mandatory = True
    vbox.pack_start(widget2, True, False, 0)

    win.show_all()
    Gtk.main()


if __name__ == "__main__":
    import sys

    # fall back to root logger for testing
    _LOG = logging.getLogger()
    sys.exit(main(sys.argv))



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2022       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext

MODULE_VERSION = "6.0"

# ------------------------------------------------------------------------
#
# Default citation formatters for Gramps
#
# ------------------------------------------------------------------------

register(
    CITE,
    id="cite-legacy",
    name=_("Legacy"),
    description=_("Legacy citation formatter"),
    version="1.0",
    gramps_target_version=MODULE_VERSION,
    status=STABLE,
    fname="legacy.py",
    authors=["The Gramps project"],
    authors_email=["http://gramps-project.org"],
)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2024       Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Base Object class for Gramps
"""

# -------------------------------------------------------------------------
#
# Python modules
#
# -------------------------------------------------------------------------
import re
from abc import ABCMeta, abstractmethod


# -------------------------------------------------------------------------
#
# BaseObject
#
# -------------------------------------------------------------------------
class BaseObject(metaclass=ABCMeta):
    """
    The BaseObject is the base class for all data objects in Gramps,
    whether primary or not.

    Its main goal is to provide common capabilites to all objects, such as
    searching through all available information.
    """

    @abstractmethod
    def serialize(self):
        """
        Convert the object to a serialized tuple of data.
        """

    @abstractmethod
    def unserialize(self, data):
        """
        Convert a serialized tuple of data to an object.
        """

    def get_object_state(self):
        """
        Get the current object state as a dictionary.

        By default this returns the public attributes of the instance.  This
        method can be overridden if the class requires other attributes or
        properties to be saved.

        This method is called to provide the information required to serialize
        the object.

        :returns: Returns a dictionary of attributes that represent the state
                  of the object.
        :rtype: dict
        """
        attr_dict = dict(
            (key, value)
            for key, value in self.__dict__.items()
            if not key.startswith("_")
        )
        attr_dict["_class"] = self.__class__.__name__
        return attr_dict

    def set_object_state(self, attr_dict):
        """
        Set the current object state using information provided in the given
        dictionary.

        By default this sets the state of the object assuming that all items in
        the dictionary map to public attributes.  This method can be overridden
        to set the state using custom functionality.  For performance reasons
        it is useful to set a property without calling its setter function.  As
        JSON provides no distinction between tuples and lists, this method can
        also be use to convert lists into tuples where required.

        :param attr_dict: A dictionary of attributes that represent the state of
                          the object.
        :type attr_dict: dict
        """
        self.__dict__.update(attr_dict)

    def matches_string(self, pattern, case_sensitive=False):
        """
        Return True if any text data in the object or any of it's child
        objects matches a given pattern.

        :param pattern: The pattern to match.
        :type pattern: str
        :param case_sensitive: Whether the match is case-sensitive.
        :type case_sensitive: bool
        :returns: Returns whether any text data in the object or any of it's
                  child objects matches a given pattern.
        :rtype: bool
        """
        # Run through its own items
        patern_upper = pattern.upper()
        for item in self.get_text_data_list():
            if not item:
                continue
            if case_sensitive:
                if item.find(pattern) != -1:
                    return True
            else:
                if item.upper().find(patern_upper) != -1:
                    return True

        # Run through child objects
        for obj in self.get_text_data_child_list():
            if obj.matches_string(pattern, case_sensitive):
                return True

        return False

    def matches_regexp(self, pattern, case_sensitive=False):
        """
        Return True if any text data in the object or any of it's child
        objects matches a given regular expression.

        :param pattern: The pattern to match.
        :type pattern: str
        :returns: Returns whether any text data in the object or any of it's
                  child objects matches a given regexp.
        :rtype: bool
        """

        # Run through its own items
        if case_sensitive:
            pattern_obj = re.compile(pattern)
        else:
            pattern_obj = re.compile(pattern, re.IGNORECASE)
        for item in self.get_text_data_list():
            if item and pattern_obj.match(item):
                return True

        # Run through child objects
        for obj in self.get_text_data_child_list():
            if obj.matches_regexp(pattern, case_sensitive):
                return True

        return False

    def get_text_data_list(self):
        """
        Return the list of all textual attributes of the object.

        :returns: Returns the list of all textual attributes of the object.
        :rtype: list
        """
        return []

    def get_text_data_child_list(self):
        """
        Return the list of child objects that may carry textual data.

        :returns: Returns the list of child objects that may carry textual data.
        :rtype: list
        """
        return []

    def get_referenced_handles(self):
        """
        Return the list of (classname, handle) tuples for all directly
        referenced primary objects.

        :returns: Returns the list of (classname, handle) tuples for referenced
                  objects.
        :rtype: list
        """
        return []

    def get_handle_referents(self):
        """
        Return the list of child objects which may, directly or through
        their children, reference primary objects.

        :returns: Returns the list of objects referencing primary objects.
        :rtype: list
        """
        return []

    def get_referenced_handles_recursively(self):
        """
        Return the list of (classname, handle) tuples for all referenced
        primary objects, whether directly or through child objects.

        :returns: Returns the list of (classname, handle) tuples for referenced
                  objects.
        :rtype: list
        """
        ret = self.get_referenced_handles()

        # Run through child objects
        for obj in self.get_handle_referents():
            ret += obj.get_referenced_handles_recursively()
        return ret

    def merge(self, acquisition):
        """
        Merge content of this object with that of acquisition.

        There are two sides to merger. First, the content of acquisition needs
        to be incorporated. Second, handles that reference acquisition (if
        there are any) need to be updated. Only the first part is handled in
        gen.lib, the second part needs access to the database and should be
        done in its own routines.

        :param acquisition: The object to incorporate.
        :type acquisition: BaseObject
        """

    @classmethod
    def create(cls, data):
        """
        Create a new instance from serialized data.
        """
        if data:
            return cls().unserialize(data)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
String mappings for constants
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..lib import Person, Citation, FamilyRelType
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext


def _T_(value, context=""):  # enable deferred translations
    return "%s\x04%s" % (context, value) if context else value


# _T_ is a gramps-defined keyword -- see po/update_po.py and po/genpot.sh

# -------------------------------------------------------------------------
#
# Integer to String mappings for constants
#
# -------------------------------------------------------------------------
gender = {
    Person.MALE: _("male"),
    Person.FEMALE: _("female"),
    Person.UNKNOWN: _("unknown", "gender"),
    Person.OTHER: _("other", "gender"),
}


def format_gender(type):
    return gender.get(type[0], _("Invalid"))


conf_strings = {
    Citation.CONF_VERY_HIGH: _T_("Very High"),
    Citation.CONF_HIGH: _T_("High"),
    Citation.CONF_NORMAL: _T_("Normal"),
    Citation.CONF_LOW: _T_("Low"),
    Citation.CONF_VERY_LOW: _T_("Very Low"),
}
# note that a list /very/ similar to this is in EditCitation._setup_fields
# but that has the glocale's translated values since it is used in the UI

family_rel_descriptions = {
    FamilyRelType.MARRIED: _(
        "A legal or common-law relationship " "between a husband and wife"
    ),
    FamilyRelType.UNMARRIED: _(
        "No legal or common-law relationship " "between man and woman"
    ),
    FamilyRelType.CIVIL_UNION: _(
        "An established relationship between " "members of the same sex"
    ),
    FamilyRelType.UNKNOWN: _("Unknown relationship between a man " "and woman"),
    FamilyRelType.CUSTOM: _("An unspecified relationship between " "a man and woman"),
}

data_recover_msg = _(
    "The data can only be recovered by Undo operation "
    "or by quitting with abandoning changes."
)



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2013 Vassilii Khachaturov <vassilii@tarunz.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""Unittest for constfunc.py"""

import unittest
from .. import constfunc

from os import environ as env


class Test_has_display(unittest.TestCase):
    def setUp(self):
        self.has = constfunc.has_display()
        self.display_nonempty = ("DISPLAY" in env) and bool(env["DISPLAY"])

    @unittest.skipUnless(constfunc.lin(), "Written for Linux only...")
    def test_consistent_with_DISPLAY_env(self):
        self.assertEqual(
            self.has,
            self.display_nonempty,
            "has_display(): {}, $DISPLAY: {}".format(
                self.has, env["DISPLAY"] if ("DISPLAY" in env) else "(unset)"
            ),
        )


if __name__ == "__main__":
    unittest.main()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2010  Michiel D. Nauta
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Provide merge capabilities for families.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..db import DbTxn
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from ..errors import MergeError
from . import MergePersonQuery


# -------------------------------------------------------------------------
#
# MergeFamilyQuery
#
# -------------------------------------------------------------------------
class MergeFamilyQuery:
    """
    Create database query to merge two families.
    """

    def __init__(self, database, phoenix, titanic, phoenix_fh=None, phoenix_mh=None):
        self.database = database
        self.phoenix = phoenix
        self.titanic = titanic
        if phoenix_fh is None:
            self.phoenix_fh = self.phoenix.get_father_handle()
        else:
            self.phoenix_fh = phoenix_fh
        if phoenix_mh is None:
            self.phoenix_mh = self.phoenix.get_mother_handle()
        else:
            self.phoenix_mh = phoenix_mh

        if self.phoenix.get_father_handle() == self.phoenix_fh:
            self.titanic_fh = self.titanic.get_father_handle()
            self.father_swapped = False
        else:
            assert self.phoenix_fh == self.titanic.get_father_handle()
            self.titanic_fh = self.phoenix.get_father_handle()
            self.father_swapped = True
        if self.phoenix.get_mother_handle() == self.phoenix_mh:
            self.titanic_mh = self.titanic.get_mother_handle()
            self.mother_swapped = False
        else:
            assert self.phoenix_mh == self.titanic.get_mother_handle()
            self.titanic_mh = self.phoenix.get_mother_handle()
            self.mother_swapped = True

    def merge_person(self, phoenix_person, titanic_person, parent, trans):
        """
        Merge two persons even if they are None; no families are merged!
        """
        new_handle = self.phoenix.get_handle()
        old_handle = self.titanic.get_handle()

        if parent == "father":
            swapped = self.father_swapped
            family_add_person_handle = (
                self.phoenix if swapped else self.titanic
            ).set_father_handle
        elif parent == "mother":
            swapped = self.mother_swapped
            family_add_person_handle = (
                self.phoenix if swapped else self.titanic
            ).set_mother_handle
        else:
            raise ValueError(_("A parent should be a father or mother."))

        if phoenix_person is None:
            if titanic_person is not None:
                raise MergeError(
                    """When merging people where one person """
                    """doesn't exist, that "person" must be the person that """
                    """will be deleted from the database."""
                )
            return
        elif titanic_person is None:
            if swapped:
                if any(
                    childref.get_reference_handle() == phoenix_person.get_handle()
                    for childref in self.phoenix.get_child_ref_list()
                ):
                    raise MergeError(
                        _(
                            "A parent and child cannot be merged. "
                            "To merge these people, you must first break the "
                            "relationship between them."
                        )
                    )

                phoenix_person.add_family_handle(new_handle)
                family_add_person_handle(phoenix_person.get_handle())
                self.database.commit_family(self.phoenix, trans)
            else:
                if any(
                    childref.get_reference_handle() == phoenix_person.get_handle()
                    for childref in self.titanic.get_child_ref_list()
                ):
                    raise MergeError(
                        _(
                            "A parent and child cannot be merged. "
                            "To merge these people, you must first break the "
                            "relationship between them."
                        )
                    )

                phoenix_person.add_family_handle(old_handle)
                family_add_person_handle(phoenix_person.get_handle())
                self.database.commit_family(self.titanic, trans)

            self.database.commit_person(phoenix_person, trans)
        else:
            query = MergePersonQuery(self.database, phoenix_person, titanic_person)
            query.execute(family_merger=False, trans=trans)

    def execute(self):
        """
        Merges two families into a single family.
        """
        new_handle = self.phoenix.get_handle()
        old_handle = self.titanic.get_handle()

        with DbTxn(_("Merge Family"), self.database) as trans:
            # commit family in case Phoenix GrampsID, relationship has changed
            self.database.commit_family(self.phoenix, trans)
            if self.phoenix_fh != self.titanic_fh:
                if self.phoenix_fh:
                    phoenix_father = self.database.get_person_from_handle(
                        self.phoenix_fh
                    )
                else:
                    phoenix_father = None
                if self.titanic_fh:
                    titanic_father = self.database.get_person_from_handle(
                        self.titanic_fh
                    )
                else:
                    titanic_father = None
                self.merge_person(phoenix_father, titanic_father, "father", trans)

            if self.phoenix_mh != self.titanic_mh:
                if self.phoenix_mh:
                    phoenix_mother = self.database.get_person_from_handle(
                        self.phoenix_mh
                    )
                else:
                    phoenix_mother = None
                if self.titanic_mh:
                    titanic_mother = self.database.get_person_from_handle(
                        self.titanic_mh
                    )
                else:
                    titanic_mother = None
                self.merge_person(phoenix_mother, titanic_mother, "mother", trans)
            # Reload families from db in case the merge_person above changed
            # them
            self.phoenix = self.database.get_family_from_handle(new_handle)
            self.titanic = self.database.get_family_from_handle(old_handle)

            if self.phoenix_fh:
                phoenix_father = self.database.get_person_from_handle(self.phoenix_fh)
            else:
                phoenix_father = None
            if self.phoenix_mh:
                phoenix_mother = self.database.get_person_from_handle(self.phoenix_mh)
            else:
                phoenix_mother = None
            self.phoenix.merge(self.titanic)
            self.database.commit_family(self.phoenix, trans)
            for childref in self.titanic.get_child_ref_list():
                child = self.database.get_person_from_handle(
                    childref.get_reference_handle()
                )
                if new_handle in child.parent_family_list:
                    child.remove_handle_references("Family", [old_handle])
                else:
                    child.replace_handle_reference("Family", old_handle, new_handle)
                self.database.commit_person(child, trans)
            if phoenix_father:
                phoenix_father.remove_family_handle(old_handle)
                self.database.commit_person(phoenix_father, trans)
            if phoenix_mother:
                phoenix_mother.remove_family_handle(old_handle)
                self.database.commit_person(phoenix_mother, trans)
            # replace the family in lds ordinances and notes
            for ref_obj, ref_handle in self.database.find_backlink_handles(
                old_handle, ["Person", "Note"]
            ):
                if ref_handle in (self.titanic_fh, self.titanic_mh):
                    continue
                obj = self.database.method("get_%s_from_handle", ref_obj)(ref_handle)
                assert obj.has_handle_reference("Family", old_handle)
                obj.replace_handle_reference("Family", old_handle, new_handle)
                if ref_handle != old_handle:
                    self.database.method("commit_%s", ref_obj)(obj, trans)

            self.database.remove_family(old_handle, trans)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2005-2007  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2009       Benny Malengier
# Copyright (C) 2010       Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# GNOME modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
import logging

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.config import config
from gramps.gui.basesidebar import BaseSidebar


# -------------------------------------------------------------------------
#
# ExpanderSidebar class
#
# -------------------------------------------------------------------------
class ExpanderSidebar(BaseSidebar):
    """
    A sidebar displaying toggle buttons and buttons with drop-down menus that
    allows the user to change the current view.
    """

    def __init__(self, dbstate, uistate, categories, views):
        self.viewmanager = uistate.viewmanager
        self.views = views

        self.expanders = []
        self.buttons = []
        self.button_handlers = []
        self.lookup = {}

        self.window = Gtk.ScrolledWindow()
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.window.add(vbox)
        self.window.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC)
        self.window.show()

        use_text = config.get("interface.sidebar-text")
        for cat_num, cat_name, cat_icon in categories:
            expander = Gtk.Expander()
            self.expanders.append(expander)

            catbox = Gtk.Box()
            image = Gtk.Image()
            image.set_from_icon_name(cat_icon, Gtk.IconSize.BUTTON)
            catbox.pack_start(image, False, False, 4)
            if use_text:
                label = Gtk.Label(label=cat_name)
                catbox.pack_start(label, False, True, 4)
            catbox.set_tooltip_text(cat_name)
            expander.set_label_widget(catbox)

            viewbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
            for view_num, view_name, view_icon in views[cat_num]:
                # create the button and add it to the sidebar
                button = self.__make_sidebar_button(
                    use_text, cat_num, view_num, view_name, view_icon
                )

                viewbox.pack_start(button, False, False, 0)
            expander.add(viewbox)
            vbox.pack_start(expander, False, True, 0)

            # Enable view switching during DnD
            # catbox.drag_dest_set(0, [], 0)
            # catbox.connect('drag_motion', self.cb_switch_page_on_dnd, cat_num)

        vbox.show_all()

    def get_top(self):
        """
        Return the top container widget for the GUI.
        """
        return self.window

    def view_changed(self, cat_num, view_num):
        """
        Called when the active view is changed.
        """
        # Expand category
        if cat_num is not None:
            self.expanders[cat_num].set_expanded(True)
        # Set new button as selected
        button_num = self.lookup.get((cat_num, view_num))
        self.__handlers_block()
        for index, button in enumerate(self.buttons):
            if index == button_num:
                button.set_active(True)
            else:
                button.set_active(False)
        self.__handlers_unblock()

    def __handlers_block(self):
        """
        Block signals to the buttons to prevent spurious events.
        """
        for idx, button in enumerate(self.buttons):
            button.handler_block(self.button_handlers[idx])

    def __handlers_unblock(self):
        """
        Unblock signals to the buttons.
        """
        for idx, button in enumerate(self.buttons):
            button.handler_unblock(self.button_handlers[idx])

    def cb_view_clicked(self, radioaction, current, cat_num):
        """
        Called when a button causes a view change.
        """
        view_num = radioaction.get_current_value()
        self.viewmanager.goto_page(cat_num, view_num)

    def __category_clicked(self, button, cat_num):
        """
        Called when a category button is clicked.
        """
        # Make the button active.  If it was already active the category will
        # not change.
        button.set_active(True)
        self.viewmanager.goto_page(cat_num, None)

    def __view_clicked(self, button, cat_num, view_num):
        """
        Called when a view button is clicked.
        """
        self.viewmanager.goto_page(cat_num, view_num)

    def cb_menu_clicked(self, menuitem, cat_num, view_num):
        """
        Called when a view is selected from a drop-down menu.
        """
        self.viewmanager.goto_page(cat_num, view_num)

    def __make_sidebar_button(self, use_text, cat_num, view_num, view_name, view_icon):
        """
        Create the sidebar button. The page_title is the text associated with
        the button.
        """
        top = Gtk.Box()

        # create the button
        button = Gtk.ToggleButton()
        button.set_relief(Gtk.ReliefStyle.NONE)
        self.buttons.append(button)
        self.lookup[(cat_num, view_num)] = len(self.buttons) - 1

        # add the tooltip
        button.set_tooltip_text(view_name)

        # connect the signal, along with the index as user data
        handler_id = button.connect("clicked", self.__view_clicked, cat_num, view_num)
        self.button_handlers.append(handler_id)
        button.show()

        # add the image. If we are using text, use the BUTTON (larger) size.
        # otherwise, use the smaller size
        hbox = Gtk.Box()
        hbox.show()
        image = Gtk.Image()
        if use_text:
            image.set_from_icon_name(view_icon, Gtk.IconSize.BUTTON)
        else:
            image.set_from_icon_name(view_icon, Gtk.IconSize.DND)
        image.show()
        hbox.pack_start(image, False, False, 0)
        hbox.set_spacing(4)

        # add text if requested
        if use_text:
            label = Gtk.Label(label=view_name)
            label.show()
            hbox.pack_start(label, False, True, 0)

        button.add(hbox)

        top.pack_start(button, False, True, 0)

        return top

    def cb_switch_page_on_dnd(self, widget, context, xpos, ypos, time, page_no):
        """
        Switches the page based on drag and drop.
        """
        self.__handlers_block()
        if self.viewmanager.notebook.get_current_page() != page_no:
            self.viewmanager.notebook.set_current_page(page_no)
        self.__handlers_unblock()



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2024       Cameron Davidson
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
A utility to make a best guess if a person is alive.  This is used to provide
privacy in reports and exports.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import logging

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..display.name import displayer as name_displayer
from ..lib.date import Date, Today
from ..lib.person import Person
from ..lib.json_utils import DataDict
from ..errors import DatabaseError
from ..const import GRAMPS_LOCALE as glocale
from ..proxy.proxybase import ProxyDbBase

LOG = logging.getLogger(".gen.utils.alive")

_ = glocale.translation.sgettext
ngettext = glocale.translation.ngettext

# DEBUGLEVEL should be 1 for production, as higher levels are rather cryptic unless
# viewed in the context of the source code.
DEBUGLEVEL = 1  # 4 = everything; 3 much detail; 2= minor detail; 1 = summary
# -------------------------------------------------------------------------
#
# Constants from config .ini keys
#
# -------------------------------------------------------------------------
# cache values; use refresh_constants() if they change
try:
    from ..config import config

    _MAX_AGE_PROB_ALIVE = config.get("behavior.max-age-prob-alive")
    _MAX_SIB_AGE_DIFF = config.get("behavior.max-sib-age-diff")
    _AVG_GENERATION_GAP = config.get("behavior.avg-generation-gap")
    _MIN_GENERATION_YEARS = config.get("behavior.min-generation-years")
except ImportError:
    # Utils used as module not part of GRAMPS
    _MAX_AGE_PROB_ALIVE = 110
    _MAX_SIB_AGE_DIFF = 20
    _AVG_GENERATION_GAP = 20
    _MIN_GENERATION_YEARS = 13


# -------------------------------------------------------------------------
#
# ProbablyAlive class
#
# -------------------------------------------------------------------------
class ProbablyAlive:
    """
    An object to hold the parameters for considering someone alive.
    """

    def __init__(
        self,
        db,
        max_sib_age_diff=None,
        max_age_prob_alive=None,
        avg_generation_gap=None,
        min_generation_years=None,
    ):
        self.db = db
        if max_sib_age_diff is None:
            max_sib_age_diff = _MAX_SIB_AGE_DIFF
        if max_age_prob_alive is None:
            max_age_prob_alive = _MAX_AGE_PROB_ALIVE
        if avg_generation_gap is None:
            avg_generation_gap = _AVG_GENERATION_GAP
        if min_generation_years is None:
            min_generation_years = _MIN_GENERATION_YEARS
        self.MAX_SIB_AGE_DIFF = max_sib_age_diff
        self.MAX_AGE_PROB_ALIVE = max_age_prob_alive
        self.AVG_GENERATION_GAP = avg_generation_gap
        self.MIN_GENERATION_YEARS = min_generation_years
        self.pset = set()

    def probably_alive_range(self, person, is_spouse=False, immediate_fam_only=False):
        """
        Find likely birth and death date ranges, either from dates of actual
        events recorded in the db or else estimating range limits from
        other events in their lives or those of close family.
        If is_spouse is True then we are calling this recursively for the
        spouse of the original "person". That will be done in two passes:
        if immediate_fam_only is True then only immediate family of spouse
        is checked, otherwise a full check is done.

        Returns: (birth_date, death_date, explain_text, related_person)
        """
        # where appropriate, some derived dates are expressed as a range.
        if person is None:
            return (None, None, "", None)
        self.pset = set()
        birth_date = None
        death_date = None
        known_to_be_dead = False
        # whether the birth evidence (date or fallback) is known for this person
        birth_evidence_direct = False
        min_birth_year = None
        max_birth_year = None
        min_birth_year_from_death = None  # values derived from 110 year extrapolations
        max_birth_year_from_death = None
        # these min/max parameters are simply years
        sib_birth_min, sib_birth_max = (None, None)
        explain_birth_min = ""
        explain_birth_max = ""
        explain_death = ""

        def get_person_bd(class_or_handle):
            """
                Looks up birth and death events for referenced person,
                using fallback dates if necessary.
                The dates will always be either None or valid values, avoiding EMPTYs.
                Only actual recorded dates are returned - there are no inferred
                limit values supplied for missing dates.

            returns  (birth_date, death_date, death_found, explain_birth, explain_death)
                                 for the referenced person
            """
            birth_date = None
            death_date = None
            death_found = False
            explain_birth = ""
            explain_death = ""

            if not class_or_handle:
                return (
                    birth_date,
                    death_date,
                    death_found,
                    explain_birth,
                    explain_death,
                )

            if isinstance(class_or_handle, (Person, DataDict)):
                thisperson = class_or_handle
            elif isinstance(class_or_handle, str):
                thisperson = self.db.get_person_from_handle(class_or_handle)
            else:
                thisperson = None

            if not thisperson:
                LOG.debug("    get_person_bd(): null person called")
                return (
                    birth_date,
                    death_date,
                    death_found,
                    explain_birth,
                    explain_death,
                )
            # is there an actual death record?  Even if yes, there may be no date,
            # in which case the EMPTY date is reported for the event.
            death_ref = thisperson.get_death_ref()
            if death_ref and death_ref.get_role().is_primary():
                evnt = self.db.get_event_from_handle(death_ref.ref)
                if evnt:
                    death_found = True
                    dateobj = evnt.get_date_object()
                    if dateobj and dateobj.is_valid():
                        death_date = dateobj
                        explain_death = _("date")

            # at this stage death_date is None or a valid date.
            # death_found is true if thisperson is known to be dead,
            #        whether or not a date was found.
            # If we have no death_date then look for fallback event such as Burial.
            # These fallbacks are fairly good indications that someone's not alive.
            # If the fallback death event does not have a valid date then it means
            # we know they are dead but not when they died.
            # So keep checking in case we get a date.
            if not death_date:
                for ev_ref in thisperson.get_primary_event_ref_list():
                    if ev_ref:
                        evnt = self.db.get_event_from_handle(ev_ref.ref)
                        if evnt and evnt.type.is_death_fallback():
                            death_date_fb = evnt.get_date_object()
                            death_found = True
                            if death_date_fb.is_valid():
                                death_date = death_date_fb
                                explain_death = _("date fallback")
                                if death_date.get_modifier() == Date.MOD_NONE:
                                    death_date.set_modifier(Date.MOD_BEFORE)
                                break  # we found a valid date, stop looking.
            # At this point:
            # * death_found is False: (no death indication found); or
            # * death_found is True. (death confirmed somehow);  In which case:
            #       * (death_date is valid) some form of death date found; or
            #       * (death_date is None and no date was recorded)
            # now repeat, looking for birth date
            birth_ref = thisperson.get_birth_ref()
            if birth_ref and birth_ref.get_role().is_primary():
                evnt = self.db.get_event_from_handle(birth_ref.ref)
                if evnt:
                    dateobj = evnt.get_date_object()
                    if dateobj and dateobj.is_valid():
                        birth_date = dateobj
                        explain_birth = _("date")

            # to here:
            #   birth_date is None: either no birth record or else no date reported; or
            #   birth_date is a valid date
            # Look for Baptism, etc events.
            # These are fairly good indications of someone's birth date.
            if not birth_date:
                for ev_ref in thisperson.get_primary_event_ref_list():
                    evnt = self.db.get_event_from_handle(ev_ref.ref)
                    if evnt and evnt.type.is_birth_fallback():
                        birth_date_fb = evnt.get_date_object()
                        if birth_date_fb and birth_date_fb.is_valid():
                            birth_date = birth_date_fb
                            explain_birth = _("date fallback")
                            break
            if DEBUGLEVEL > 3:
                LOG.debug(
                    "           << get_person_bd for [%s], birth %s, death %s",
                    thisperson.get_gramps_id(),
                    birth_date,
                    death_date,
                )
            return (birth_date, death_date, death_found, explain_birth, explain_death)

        birth_date, death_date, known_to_be_dead, explain_birth_min, explain_death = (
            get_person_bd(person)
        )

        if birth_date is not None:
            birth_evidence_direct = True
            if death_date is not None:
                explanation = _(
                    "Direct evidence for this person - birth: {birth_min_src}, death: {death_src}"
                ).format(
                    birth_min_src=explain_birth_min,
                    death_src=explain_death,
                )
                return (
                    birth_date,
                    death_date,
                    explanation,
                    person,
                )  # direct self evidence

        # birth and/or death dates are not known, so let's see what we can estimate.
        # First: minimum is X years before death;
        # Second: get the parent's birth/death dates if available, so we can constrain
        # to sensible values - mother's age and parent's death.
        # Finally: get birth dates for any full siblings to further constrain.
        # Currently only look at full siblings, ranges would get wider for half sibs.

        if birth_date is None:
            # only need to estimate birth_date if we have no more direct evidence.
            if death_date is not None:
                # person died so guess initial limits to birth date
                if death_date.get_year_valid():
                    max_birth_year_from_death = death_date.get_year()
                    min_birth_year_from_death = (
                        max_birth_year_from_death - self.MAX_AGE_PROB_ALIVE
                    )

            m_birth = m_death = None  # mother's birth and death dates
            f_birth = f_death = None  # father's
            parents = None  # Family with parents
            parenth_p1 = person.get_main_parents_family_handle()
            if parenth_p1:
                parents = self.db.get_family_from_handle(parenth_p1)
                mother_handle_p1 = parents.get_mother_handle()
                m_birth, m_death = get_person_bd(mother_handle_p1)[0:2]
                father_handle_p1 = parents.get_father_handle()
                f_birth, f_death = get_person_bd(father_handle_p1)[0:2]
            # now scan siblings
            family_list = person.get_parent_family_handle_list()
            for family_handle in family_list:
                family = self.db.get_family_from_handle(family_handle)
                if family is None:
                    continue
                if parents is not None and family_handle != parenth_p1:
                    LOG.debug(
                        "      skipping family %s but parents is %s.",
                        family.get_gramps_id(),
                        parents.get_gramps_id(),
                    )
                    continue
                for child_ref in family.get_child_ref_list():
                    child_handle = child_ref.ref
                    child = self.db.get_person_from_handle(child_handle)
                    if child is None or child == person:
                        continue
                    need_birth_fallback = True
                    # Go through once looking for direct evidence:
                    # extract the range of birth dates, either direct or fallback
                    for ev_ref in child.get_primary_event_ref_list():
                        evnt = self.db.get_event_from_handle(ev_ref.ref)
                        if evnt and evnt.type.is_birth():
                            dobj = evnt.get_date_object()
                            if dobj and dobj.get_year_valid():
                                year = dobj.get_year()
                                need_birth_fallback = False
                                if sib_birth_min is None or year < sib_birth_min:
                                    sib_birth_min = year
                                if sib_birth_max is None or year > sib_birth_max:
                                    sib_birth_max = year
                    # scan event list again looking for fallback:
                    if need_birth_fallback:
                        for ev_ref in child.get_primary_event_ref_list():
                            evnt = self.db.get_event_from_handle(ev_ref.ref)
                            if evnt and evnt.type.is_birth_fallback():
                                dobj = evnt.get_date_object()
                                if dobj and dobj.get_year_valid():
                                    # if sibling birth date too far away, then
                                    # cannot be alive:
                                    year = dobj.get_year()
                                    if sib_birth_min is None or year < sib_birth_min:
                                        sib_birth_min = year
                                    if sib_birth_max is None or year > sib_birth_max:
                                        sib_birth_max = year
            # Now combine estimates based on parents and siblings:
            # Make sure child is born after both parents are old enough
            if m_birth:
                min_birth_year = m_birth.get_year() + self.MIN_GENERATION_YEARS
                explain_birth_min = _("mother's age")
            if f_birth:
                min_from_f = f_birth.get_year() + self.MIN_GENERATION_YEARS
                if min_birth_year is None or min_from_f > min_birth_year:
                    min_birth_year = min_from_f
                    explain_birth_min = _("father's age")
            if min_birth_year_from_death:
                if min_birth_year is None or min_birth_year_from_death > min_birth_year:
                    min_birth_year = min_birth_year_from_death
                    explain_birth_min = _("from death date")
            # Calculate the latest year that the child could have been born
            if m_death:
                max_birth_year = m_death.get_year()
                explain_birth_max = _("mother's death")
            if f_death:
                max_from_f = f_death.get_year() + 1
                if max_birth_year is None or max_from_f < max_birth_year:
                    max_birth_year = max_from_f
                    explain_birth_max = _("father's death")
            if max_birth_year_from_death:
                if max_birth_year is None or max_birth_year_from_death < max_birth_year:
                    max_birth_year = max_birth_year_from_death
                    explain_birth_max = _("person's death")

            # sib_xx_min/max are either both None or both have a value (maybe the same)
            if sib_birth_max:
                min_from_sib = sib_birth_max - self.MAX_SIB_AGE_DIFF
                if min_birth_year is None or min_from_sib > min_birth_year:
                    min_birth_year = min_from_sib
                    explain_birth_min = _("oldest sibling's age")

                max_from_sib = sib_birth_min + self.MAX_SIB_AGE_DIFF
                if max_birth_year is None or max_from_sib < max_birth_year:
                    max_birth_year = max_from_sib
                    explain_birth_max = _("youngest sibling's age")

        if birth_date is None or not birth_date.is_valid():
            birth_date = Date()  # make sure we have an empty date
            #  use proxy estimate
            if min_birth_year and max_birth_year:
                # create a range set
                birth_range = list(Date.EMPTY + Date.EMPTY)
                birth_range[Date._POS_YR] = min_birth_year
                birth_range[Date._POS_RYR] = max_birth_year
                birth_date.set(modifier=Date.MOD_RANGE, value=tuple(birth_range))
            else:
                if min_birth_year:
                    birth_date.set_yr_mon_day(min_birth_year, 1, 1)
                    birth_date.set_modifier(Date.MOD_AFTER)
                elif max_birth_year:
                    birth_date.set_yr_mon_day(max_birth_year, 12, 31)
                    birth_date.set_modifier(Date.MOD_BEFORE)
            birth_date.recalc_sort_value()

        # If we have no death date but we know death has happened then
        # we set death range somewhere between birth and yesterday.
        # otherwise we assume MAX years after birth
        if death_date is None:
            if birth_date and birth_date.is_valid():
                death_date = Date(birth_date)
                max_death_date = birth_date.copy_offset_ymd(
                    year=self.MAX_AGE_PROB_ALIVE
                )
                if known_to_be_dead:
                    if max_death_date.match(Today(), ">="):
                        max_death_date = Today()
                        max_death_date.set_yr_mon_day_offset(
                            day=-1
                        )  # make it yesterday
                    # range start value stays at birth date
                    death_date.set_modifier(Date.MOD_RANGE)
                    death_date.set_text_value("")
                    death_date.set2_yr_mon_day(
                        max_death_date.get_year(),
                        max_death_date.get_month(),
                        max_death_date.get_day(),
                    )
                    explain_death = _("birth date and known to be dead")
                else:
                    death_date.set_yr_mon_day_offset(year=self.MAX_AGE_PROB_ALIVE)
                    explain_death = _("offset from birth date")
                death_date.recalc_sort_value()
            else:
                death_date = Date()

        # at this stage we should have valid dates for both birth and death,
        #  or else both are zero (if None then it's a bug).
        if birth_evidence_direct:
            explanation = _(
                "Direct evidence for this person - birth: {birth_min_src}, death: {death_src}"
            ).format(
                birth_min_src=explain_birth_min,
                death_src=explain_death,
            )
        elif explain_birth_max == "":
            explanation = _(
                "inferred from parents and siblings - birth: {birth_min_src}, death: {death_src}"
            ).format(
                birth_min_src=explain_birth_min,
                death_src=explain_death,
            )
        else:
            explanation = _(
                "inferred from parents and siblings - birth: {birth_min_src} (earliest) to {birth_max_src} (latest), death: {death_src}"
            ).format(
                birth_min_src=explain_birth_min,
                birth_max_src=explain_birth_max,
                death_src=explain_death,
            )

        if birth_date.is_valid() and death_date.is_valid():
            return (birth_date, death_date, explanation, person)

        # have finished immediate family, so try spouse, as the
        # remaining person (probably) of this generation ..

        def spouse_test(only_immediate_family):
            # test against spouse details - this is done in two passes, at different
            # stages of generating dates for the reference person.:
            # 1. test spouse details only - this should be a reasonable proxy for
            # reference person, after immediate family.
            # 2. run full test recursing into probably_alive_range
            # Only run this pass at the end - tests have higher uncertainty than the
            # same test on the reference person.
            # We allow for an age difference +/- AVG_GENERATION_GAP
            # which, assuming defaults, results in 150 year "probably alive" range.
            # In reality, if we have reached this far then any value is unreliable.

            LOG.debug(
                "    ----- trying spouse check: %s",
                "immediate family only" if only_immediate_family else "full tree",
            )
            for family_handle in person.get_family_handle_list():
                family = self.db.get_family_from_handle(family_handle)
                if family:
                    mother_handle = family.get_mother_handle()
                    father_handle = family.get_father_handle()
                    if mother_handle is None or father_handle is None:
                        if DEBUGLEVEL > 1:
                            LOG.debug(
                                "         single parent family: [%s]",
                                family.get_gramps_id(),
                            )
                        # no recorded spouse
                        continue
                    spouse = None
                    if mother_handle == person.handle:
                        spouse = self.db.get_person_from_handle(father_handle)
                    elif father_handle == person.handle:
                        spouse = self.db.get_person_from_handle(mother_handle)
                    if spouse is not None:
                        date1, date2, explain, other = self.probably_alive_range(
                            spouse,
                            is_spouse=True,
                            immediate_fam_only=only_immediate_family,
                        )
                        if DEBUGLEVEL > 2:
                            LOG.debug(
                                "            found spouse [%s], returned b:%s, d:%s, because:%s",
                                spouse.get_gramps_id(),
                                date1,
                                date2,
                                explain,
                            )
                        if date1 and date1.get_year() != 0:
                            birth_date = date1.copy_offset_ymd(-self.AVG_GENERATION_GAP)
                            if birth_date.is_compound():
                                # it will have already offset both values, so correct that
                                # and then offset to be 1 GEN GAP higher.
                                birth_date.set2_yr_mon_day_offset(
                                    2 * self.AVG_GENERATION_GAP
                                )
                            else:
                                birth_date.set_modifier(Date.MOD_RANGE)
                                birth_date.set_text_value("")
                                # duplicate lower birth limit
                                birth_date.set2_yr_mon_day(
                                    date1.get_year(),
                                    date1.get_month(),
                                    date1.get_day(),
                                )
                                # then extend upper limit the other direction
                                birth_date.set2_yr_mon_day_offset(
                                    self.AVG_GENERATION_GAP
                                )
                            death_date = birth_date.copy_offset_ymd(
                                self.MAX_AGE_PROB_ALIVE
                            )

                            return (
                                birth_date,
                                death_date,
                                _("a spouse's birth-related date, ") + explain,
                                other,
                            )
                        if date2 and date2.get_year() != 0:
                            return (
                                Date().copy_ymd(
                                    date2.get_year()
                                    + self.AVG_GENERATION_GAP
                                    - self.MAX_AGE_PROB_ALIVE
                                ),
                                Date().copy_ymd(
                                    date2.get_year() + self.AVG_GENERATION_GAP
                                ),
                                _("a spouse's death-related date, ") + explain,
                                other,
                            )

                    # Let's check the family events and see if we find something
                    for ref in family.get_event_ref_list():
                        if ref:
                            event = self.db.get_event_from_handle(ref.ref)
                            if event:
                                date = event.get_date_object()
                                year = date.get_year()
                                if year != 0:
                                    other = None
                                    if person.handle == mother_handle and father_handle:
                                        other = self.db.get_person_from_handle(
                                            father_handle
                                        )
                                    elif (
                                        person.handle == father_handle and mother_handle
                                    ):
                                        other = self.db.get_person_from_handle(
                                            mother_handle
                                        )
                                    return (
                                        Date().copy_ymd(year - self.AVG_GENERATION_GAP),
                                        Date().copy_ymd(
                                            year
                                            - self.AVG_GENERATION_GAP
                                            + self.MAX_AGE_PROB_ALIVE
                                        ),
                                        _("event with spouse"),
                                        other,
                                    )
            return (None, None, "", None)

        if not is_spouse:
            birth_date, death_date, explain, who = spouse_test(True)
            if birth_date is not None and death_date is not None:
                return (birth_date, death_date, explain, who)
        elif immediate_fam_only:
            return (None, None, "", None)

        # Try to estimate probable lifespan by scanning descendants

        def recurse_descendants(person, generation):
            """
            Recursively scan descendants' tree to determine likely birth/death
            dates for the person specified.
            Returns the range of birth and/or deaths years for the closest generation
            in which any are available.
            generation: gets incremented as we descend the tree.
            Returns: birth_year_min, birth_year_max,
                death_year_min, death_year_max,
                n_generations,
                child
            min and max years will be either both None or both valid values
            If all years are None then n_generations will be None
            """

            no_valid_descendant = (None, None, None, None, None, None)
            if person.handle in self.pset:
                LOG.debug(
                    "....... person %s skipped - already seen in descendants test",
                    person.get_gramps_id(),
                )
                return no_valid_descendant
            if DEBUGLEVEL > 2:
                LOG.debug(
                    "     %s recursing into person [%s] %s, gen %s",
                    "..." * generation,
                    person.get_gramps_id(),
                    person.get_primary_name().get_gedcom_name(),
                    generation,
                )
            self.pset.add(person.handle)
            child_result = list()
            birth_min = birth_max = None
            death_min = death_max = None
            for family_handle in person.get_family_handle_list():
                # only families in which person is a parent or spouse of parent
                family = self.db.get_family_from_handle(family_handle)
                if not family:
                    # can happen with LivingProxyDb(PrivateProxyDb(db))
                    continue
                for child_ref in family.get_child_ref_list():
                    child_handle = child_ref.ref
                    bdate, ddate, dfound, expb, expd = get_person_bd(child_handle)
                    cd = dict(
                        handle=child_handle,
                        birthdate=bdate,
                        deathdate=ddate,
                        deathfound=dfound,
                        birth_expl=expb,
                        death_expl=expd,
                    )
                    child_result.append(cd)
                    if bdate is not None or ddate is not None:
                        if bdate is not None:
                            byear = bdate.get_year()
                            if birth_min is None or byear < birth_min:
                                birth_min = byear
                            if birth_max is None or byear > birth_max:
                                birth_max = byear
                        if ddate is not None:
                            dyear = ddate.get_year()
                            if death_min is None or dyear < death_min:
                                death_min = dyear
                            if death_max is None or dyear > death_max:
                                death_max = dyear

            # if we have something at this stage then just report it and descend no further
            if birth_min is not None or death_min is not None:
                return (birth_min, birth_max, death_min, death_max, generation, person)
            # otherwise recursively scan childrens' descendants, accumulating the results
            nextgen = list()
            mingen = None
            for childdict in child_result:
                child_handle = childdict["handle"]
                child = self.db.get_person_from_handle(child_handle)

                bmin, bmax, dmin, dmax, ngens, who = recurse_descendants(
                    child, 1 + generation
                )
                if ngens is not None:
                    if mingen is None or ngens < mingen:
                        mingen = ngens
                    ngd = dict(
                        bmin=bmin,
                        bmax=bmax,
                        dmin=dmin,
                        dmax=dmax,
                        ngens=ngens,
                        who=who,
                    )
                    nextgen.append(ngd)
            # having accumulated all results from this generation's descendants,
            # identify the values from the closest generation
            if mingen is None:
                return no_valid_descendant
            gen_bmin = gen_bmax = None  # generational min/max birth years
            gen_dmin = gen_dmax = None  # generational min/max death years
            retb_who = retd_who = None
            for ngd in nextgen:
                if ngd["ngens"] > mingen:
                    continue
                bmin = ngd["bmin"]
                dmin = ngd["dmin"]
                if bmin is not None:
                    if gen_bmin is None or bmin < gen_bmin:
                        gen_bmin = bmin
                        retb_who = ngd["who"]
                    bmax = ngd["bmax"]
                    if gen_bmax is None or bmax > gen_bmax:
                        gen_bmax = bmax
                if dmin is not None:
                    if gen_dmin is None or dmin < gen_dmin:
                        gen_dmin = dmin
                    dmax = ngd["dmax"]
                    if gen_dmax is None or dmax > gen_dmax:
                        gen_dmax = dmax
                        retd_who = ngd["who"]

            # at this stage we have summary of closest result from all descendants of person
            if gen_bmin is not None or gen_dmin is not None:
                return (
                    gen_bmin,
                    gen_bmax,
                    gen_dmin,
                    gen_dmax,
                    mingen,
                    retd_who if retb_who is None else retb_who,
                )

            return no_valid_descendant

        def estimate_bd_range_from_descendants(person):
            """
            wrapper function to initiate descendant recursion and process final results.
            """
            bmin, bmax, dmin, dmax, ngens, other = recurse_descendants(person, 1)

            date1, date2, explain = None, None, ""
            if bmin is not None:
                # This could be extended to create more realistic ranges, but, let's face it,
                # if these dates are important then the users should be making realistic estimates
                # themselves.  We'll just use averages...
                meanbirth = (bmin + bmax) / 2
                if DEBUGLEVEL > 2:
                    LOG.debug(
                        "     == desc for %s returned bmin:%s, bmax:%s, ngens:%s, dmin:%s, dmax:%s, who:%s",
                        person.get_gramps_id(),
                        bmin,
                        bmax,
                        ngens,
                        dmin,
                        dmax,
                        (
                            "None"
                            if other is None
                            else other.get_primary_name().get_gedcom_name()
                        ),
                    )
                birth_year = int(meanbirth - (ngens * self.AVG_GENERATION_GAP))
                date1 = Date()
                date1.set_yr_mon_day(birth_year, 1, 1)
                date1.set_modifier(Date.MOD_ABOUT)
                date2 = date1.copy_offset_ymd(self.MAX_AGE_PROB_ALIVE)
                explain = ngettext(
                    "descendant birth: {number_of} generation ",
                    "descendant birth: {number_of} generations ",
                    ngens,
                ).format(number_of=ngens)
            elif dmin is not None:
                # no births, just death dates ... unreliable estimates only
                # An upper limit would be based on min_generation_gap below the first death.
                # A lower limit could be based on no child exceeding 110 year
                upper_birth_year = dmin - (ngens * self.MIN_GENERATION_YEARS)
                lower_birth_year = dmax - (
                    ngens * self.AVG_GENERATION_GAP + self.MAX_AGE_PROB_ALIVE
                )
                if lower_birth_year > upper_birth_year:
                    upper_birth_year, lower_birth_year = (
                        lower_birth_year,
                        upper_birth_year,
                    )
                date1 = Date()
                date1.set_yr_mon_day(lower_birth_year, 1, 1)
                date1.set_modifier(Date.MOD_RANGE)
                date1.set2_yr_mon_day(upper_birth_year, 1, 1)
                date2 = date1.copy_offset_ymd(self.MAX_AGE_PROB_ALIVE)
                explain = ngettext(
                    "descendant death: {number_of} generation ",
                    "descendant death: {number_of} generations ",
                    ngens,
                ).format(number_of=ngens)
            if date1 and date2:
                return (date1, date2, explain, other)
            return (None, None, "", None)

        LOG.debug("    ------- checking descendants of [%s]", person.get_gramps_id())
        date1, date2, explain, other = None, None, "", None
        try:
            date1, date2, explain, other = estimate_bd_range_from_descendants(person)

        except RuntimeError:
            raise DatabaseError(
                _("Database error: loop in descendants of %s.")
                % name_displayer.display(person)
            )

        if date1 and date2:
            return (date1, date2, explain, other)

        self.pset = set()  # clear the list from descendant check

        # Try to estimate probable lifespan by scanning ancestors. We have already
        # checked person's parents, so we should scan each of their ancestors

        def estimate_bd_range_from_ancestors(person, year, generation):
            """
            Estimate birth and death year ranges based on a person's ancestors.
            The year value is the average number of years between generations.
            generation parameter is current depth of recursion.
            """
            range_not_found = (None, None, "", None, None)
            if person.handle in self.pset:
                LOG.debug(
                    "....... person %s skipped - already seen in ancestor test",
                    person.get_gramps_id(),
                )
                return range_not_found
            self.pset.add(person.handle)

            family_handle = person.get_main_parents_family_handle()
            if family_handle:
                family = self.db.get_family_from_handle(family_handle)
                if not family:
                    # can happen with LivingProxyDb(PrivateProxyDb(db))
                    return range_not_found
                mother_handle = family.get_mother_handle()
                (
                    mother_birth,
                    mother_death,
                    mother_death_found,
                    mother_expl_b,
                    mother_expl_d,
                ) = get_person_bd(mother_handle)

                father_handle = family.get_father_handle()
                (
                    father_birth,
                    father_death,
                    father_death_found,
                    father_expl_b,
                    father_expl_d,
                ) = get_person_bd(father_handle)

                parent_birth = mother_birth
                explan = _("mother birth ") + mother_expl_b
                parenth = mother_handle
                if parent_birth is None:
                    parent_birth = father_birth
                    explan = _("father birth ") + father_expl_b
                    parenth = father_handle
                elif father_birth is not None:
                    # have both births, try for youngest
                    if father_birth.match(mother_birth, ">"):
                        parent_birth = father_birth
                        explan = _("father birth ") + father_expl_b
                        parenth = father_handle
                if parent_birth is not None:
                    person_birth = parent_birth.copy_offset_ymd(year)
                    person_death = person_birth.copy_offset_ymd(self.MAX_AGE_PROB_ALIVE)
                    return (
                        person_birth,
                        person_death,
                        _("ancestor ") + explan,
                        self.db.get_person_from_handle(parenth),
                        generation,
                    )
                # no useful birth, try death...
                first_parent_death = parent_death = mother_death
                explan = _("mother death ") + mother_expl_d
                parenth = mother_handle
                if parent_death is None:
                    first_parent_death = parent_death = father_death
                    explan = _("father death ") + father_expl_d
                    parenth = father_handle
                elif father_death is not None:
                    # have both deaths, try for last to die
                    if father_death.match(mother_death, ">"):
                        parent_death = father_death
                        explan = _("father death ") + father_expl_b
                        parenth = father_handle
                        first_parent_death = mother_death
                if parent_death is not None:
                    person_birth = parent_death.copy_offset_ymd(
                        year - self.MAX_AGE_PROB_ALIVE
                    )
                    person_birth.set_modifier(Date.MOD_RANGE)
                    person_birth.set2_yr_mon_day(
                        year + first_parent_death.get_year(), 1, 1
                    )
                    person_death = person_birth.copy_offset_ymd(self.MAX_AGE_PROB_ALIVE)
                    return (
                        person_birth,
                        person_death,
                        _("ancestor ") + explan,
                        self.db.get_person_from_handle(parenth),
                        generation,
                    )

                # nothing found yet, so recurse up the mother's line first
                # This becomes a depth-first search, which is undesirable,
                # but we choose the shortest number of generations from the responses
                # not very efficient, but...
                gen_m = gen_f = None
                if mother_handle is not None:
                    date1_m, date2_m, explan_m, other_m, gen_m = (
                        estimate_bd_range_from_ancestors(
                            self.db.get_person_from_handle(mother_handle),
                            year + self.AVG_GENERATION_GAP,
                            generation + 1,
                        )
                    )
                # now try the father's line
                if father_handle is not None:
                    date1_f, date2_f, explan_f, other_f, gen_f = (
                        estimate_bd_range_from_ancestors(
                            self.db.get_person_from_handle(father_handle),
                            year + self.AVG_GENERATION_GAP,
                            generation + 1,
                        )
                    )
                # now decide which of maternal/paternal lines is better choice.
                use_side = "none"
                if gen_m is not None and gen_f is not None:
                    # if both maternal and paternal ancestral lines returned values,
                    # then take shortest depth.
                    if gen_f < gen_m:
                        use_side = "father"
                    else:
                        use_side = "mother"
                elif gen_m is not None:
                    use_side = "mother"
                elif gen_f is not None:
                    use_side = "father"

                if use_side == "mother":
                    if date1_m and date2_m:
                        return (date1_m, date2_m, explan_m, other_m, gen_m)
                elif use_side == "father":
                    if date1_f and date2_f:
                        return (date1_f, date2_f, explan_f, other_f, gen_f)

            return range_not_found

        if parenth_p1:
            LOG.debug("    ------ checking ancestors %s", person.get_gramps_id())
            try:
                date1, date2, explain, other, gen = estimate_bd_range_from_ancestors(
                    person, int(self.AVG_GENERATION_GAP), 1
                )
            except RuntimeError:
                raise DatabaseError(
                    _("Database error: loop in ancestors of %s.")
                    % name_displayer.display(person)
                )

            if date1 is not None and date2 is not None:
                return (date1, date2, explain, other)

        if not is_spouse:  # if you are not in recursion, let's recurse again:
            birth_date, death_date, explain, who = spouse_test(False)
            if birth_date is not None and death_date is not None:
                return (birth_date, death_date, explain, who)

        # If we can't find any reason to believe that they are dead we
        # must assume they are alive.

        return (None, None, "", None)


# -------------------------------------------------------------------------
#
# probably_alive
#
# -------------------------------------------------------------------------
def probably_alive(
    person,
    db,
    current_date=None,
    limit=0,
    max_sib_age_diff=None,
    max_age_prob_alive=None,
    avg_generation_gap=None,
    return_range=False,
):
    """
    Return true if the person may be alive on current_date.

    This works by a process of elimination. If we can't find a good
    reason to believe that someone is dead then we assume they must
    be alive.

    :param current_date: a date object that is not estimated or modified
                         (defaults to today if None or a blank string)
    :param limit: number of years to check beyond death_date
    :param max_sib_age_diff: maximum sibling age difference, in years
            if None then default to the setting in user config
    :param max_age_prob_alive: maximum age of a person, in years
            if None then default to the setting in user config
    :param avg_generation_gap: average generation gap, in years
    """
    LOG.debug(
        " *** probably_alive() called for [%s] %s: ",
        person.get_gramps_id(),
        person.get_primary_name().get_gedcom_name(),
    )
    # First, get the probable birth and death ranges for
    # this person from the real database:
    birth, death, explain, relative = probably_alive_range(
        person, db, max_sib_age_diff, max_age_prob_alive, avg_generation_gap
    )
    if current_date is None or not current_date.is_valid():
        current_date = Today()

    if DEBUGLEVEL > 0:
        if relative is None:
            rel_id = "nobody"
        else:
            rel_id = relative.get_gramps_id()
        LOG.debug(
            "      range: b.%s, d.%s vs %s reason: %s to [%s]",
            birth,
            death,
            current_date,
            explain,
            rel_id,
        )
    if not birth and not death:
        # no evidence, must consider alive
        LOG.debug(
            "      [%s] %s: decided alive - no evidence",
            person.get_gramps_id(),
            person.get_primary_name().get_gedcom_name(),
        )
        return (True, None, None, _("no evidence"), None) if return_range else True
    if not birth or not death:
        # insufficient evidence, must consider alive
        LOG.debug(
            "   LOGIC ERROR -  [%s] %s: only %s found; decided alive",
            person.get_gramps_id(),
            person.get_primary_name().get_gedcom_name(),
            "birth" if birth else "death",
        )
        return (True, None, None, _("no evidence"), None) if return_range else True
    # must have dates from here:
    if limit:
        death += limit  # add these years to death
    # Finally, check to see if current_date is between dates
    # ---true if  current_date >= birth(min)   and  true if current_date < death
    # these include true if current_date is within the estimated range
    result = current_date.match(birth, ">=") and current_date.match(death, "<")
    if DEBUGLEVEL > 3:
        (bthmin, bthmax) = birth.get_start_stop_range()
        (dthmin, dthmax) = death.get_start_stop_range()
        (dmin, dmax) = current_date.get_start_stop_range()
        LOG.debug(
            "        alive=%s, btest: %s, dtest: %s (born %s-%s, dd %s-%s) vs (%s-%s)",
            result,
            current_date.match(birth, ">="),
            current_date.match(death, "<"),
            bthmin,
            bthmax,
            dthmin,
            dthmax,
            dmin,
            dmax,
        )
    if return_range:
        return (result, birth, death, explain, relative)

    return result


def probably_alive_range(
    person, db, max_sib_age_diff=None, max_age_prob_alive=None, avg_generation_gap=None
):
    """
    Computes estimated birth and death date ranges.
    Returns: (birth_date, death_date, explain_text, related_person)
    """
    # First, find the real database to use all people
    # for determining alive status:
    basedb = db
    while isinstance(basedb, ProxyDbBase):
        basedb = basedb.db
    # Now, we create a wrapper for doing work:
    pbac = ProbablyAlive(
        basedb, max_sib_age_diff, max_age_prob_alive, avg_generation_gap
    )
    return pbac.probably_alive_range(person)


def update_constants():
    """
    Used to update the constants that are cached in this module.
    """

    global _MAX_AGE_PROB_ALIVE, _MAX_SIB_AGE_DIFF
    global _AVG_GENERATION_GAP, _MIN_GENERATION_YEARS
    _MAX_AGE_PROB_ALIVE = config.get("behavior.max-age-prob-alive")
    _MAX_SIB_AGE_DIFF = config.get("behavior.max-sib-age-diff")
    _AVG_GENERATION_GAP = config.get("behavior.avg-generation-gap")
    _MIN_GENERATION_YEARS = config.get("behavior.min-generation-years")



# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2011,2015  Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Gtk modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository.GLib import markup_escape_text

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.plug import Gramplet
from gramps.gui.widgets import Photo
from gramps.gen.utils.place import conv_lat_lon, coord_formats
from gramps.gen.utils.file import media_path_full
from gramps.gen.display.place import displayer as place_displayer
from gramps.gen.const import COLON, GRAMPS_LOCALE as glocale
from gramps.gen.config import config

_ = glocale.translation.gettext


class PlaceDetails(Gramplet):
    """
    Displays details for a place.
    """

    def init(self):
        self.gui.WIDGET = self.build_gui()
        self.gui.get_container_widget().remove(self.gui.textview)
        self.gui.get_container_widget().add(self.gui.WIDGET)

    def build_gui(self):
        """
        Build the GUI interface.
        """
        self.top = Gtk.Box()
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.photo = Photo(self.uistate.screen_height() < 1000)
        self.title = Gtk.Label(halign=Gtk.Align.START)
        self.title.set_selectable(True)
        vbox.pack_start(self.title, False, True, 7)
        self.grid = Gtk.Grid(orientation=Gtk.Orientation.VERTICAL)
        self.grid.set_column_spacing(10)
        vbox.pack_start(self.grid, False, True, 0)
        self.top.pack_start(self.photo, False, True, 5)
        self.top.pack_start(vbox, False, True, 10)
        self.top.show_all()
        return self.top

    def add_row(self, title, value):
        """
        Add a row to the table.
        """
        label = Gtk.Label(
            label=title + COLON, halign=Gtk.Align.END, valign=Gtk.Align.START
        )
        label.set_selectable(True)
        label.show()
        value = Gtk.Label(label=value, halign=Gtk.Align.START)
        value.set_selectable(True)
        value.show()
        self.grid.add(label)
        self.grid.attach_next_to(value, label, Gtk.PositionType.RIGHT, 1, 1)

    def clear_grid(self):
        """
        Remove all the rows from the grid.
        """
        list(map(self.grid.remove, self.grid.get_children()))

    def db_changed(self):
        self.connect(self.dbstate.db, "place-update", self.update)
        self.connect_signal("Place", self.update)

    def update_has_data(self):
        active_handle = self.get_active("Person")
        if active_handle:
            active_person = self.dbstate.db.get_person_from_handle(active_handle)
            self.set_has_data(active_person is not None)
        else:
            self.set_has_data(False)

    def main(self):
        self.display_empty()
        active_handle = self.get_active("Place")
        if active_handle:
            place = self.dbstate.db.get_place_from_handle(active_handle)
            self.top.hide()
            if place:
                self.display_place(place)
                self.set_has_data(True)
            else:
                self.set_has_data(False)
            self.top.show()
        else:
            self.set_has_data(False)

    def display_place(self, place):
        """
        Display details of the active place.
        """
        self.load_place_image(place)
        title = place_displayer.display(self.dbstate.db, place)
        self.title.set_markup(
            "<span size='large' weight='bold'>%s</span>" % markup_escape_text(title)
        )

        self.clear_grid()
        self.add_row(_("Name"), place.get_name().get_value())
        self.add_row(_("Type"), place.get_type())
        self.display_separator()
        self.display_alt_names(place)
        self.display_separator()
        lat, lon = conv_lat_lon(
            place.get_latitude(),
            place.get_longitude(),
            format=coord_formats[config.get("preferences.coord-format")],
        )
        if lat:
            self.add_row(_("Latitude"), lat)
        if lon:
            self.add_row(_("Longitude"), lon)

    def display_alt_names(self, place):
        """
        Display alternative names for the place.
        """
        alt_names = [
            (
                "%s (%s)" % (name.get_value(), name.get_language())
                if name.get_language()
                else name.get_value()
            )
            for name in place.get_alternative_names()
        ]
        if len(alt_names) > 0:
            self.add_row(_("Alternative Names"), "\n".join(alt_names))

    def display_empty(self):
        """
        Display empty details when no repository is selected.
        """
        self.photo.set_image(None)
        self.photo.set_uistate(None, None)
        self.title.set_text("")
        self.clear_grid()

    def display_separator(self):
        """
        Display an empty row to separate groupd of entries.
        """
        label = Gtk.Label()
        label.set_markup("<span font='sans 4'> </span>")
        label.set_selectable(True)
        label.show()
        self.grid.add(label)

    def load_place_image(self, place):
        """
        Load the primary image if it exists.
        """
        media_list = place.get_media_list()
        if media_list:
            media_ref = media_list[0]
            object_handle = media_ref.get_reference_handle()
            obj = self.dbstate.db.get_media_from_handle(object_handle)
            full_path = media_path_full(self.dbstate.db, obj.get_path())
            mime_type = obj.get_mime_type()
            if mime_type and mime_type.startswith("image"):
                self.photo.set_image(full_path, mime_type, media_ref.get_rectangle())
                self.photo.set_uistate(self.uistate, object_handle)
            else:
                self.photo.set_image(None)
                self.photo.set_uistate(None, None)
        else:
            self.photo.set_image(None)
            self.photo.set_uistate(None, None)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2011       Adam Stein <adam@csh.rit.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Image manipulation routines.
"""

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import os
import sys
import tempfile

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..const import GRAMPS_LOCALE as glocale

_ = glocale.translation.gettext


def crop_percentage_to_subpixel(width, height, crop):
    """
    Convert from Gramps cropping coordinates [0, 100] to
    pixels, given image width and height. No rounding to pixel resolution.
    """
    return (
        crop[0] / 100.0 * width,
        crop[1] / 100.0 * height,
        crop[2] / 100.0 * width,
        crop[3] / 100.0 * height,
    )


def crop_percentage_to_pixel(width, height, crop):
    return map(int, crop_percentage_to_subpixel(width, height, crop))


# -------------------------------------------------------------------------
#
# resize_to_jpeg
#
# -------------------------------------------------------------------------
def resize_to_jpeg(source, destination, width, height, crop=None):
    """
    Create the destination, derived from the source, resizing it to the
    specified size, while converting to JPEG.

    :param source: source image file, in any format that gtk recognizes
    :type source: unicode
    :param destination: destination image file, output written in jpeg format
    :type destination: unicode
    :param width: desired width of the destination image
    :type width: int
    :param height: desired height of the destination image
    :type height: int
    :param crop: cropping coordinates
    :type crop: array of integers ([start_x, start_y, end_x, end_y])
    """
    from gi.repository import GdkPixbuf

    img = GdkPixbuf.Pixbuf.new_from_file(source)

    if crop:
        (start_x, start_y, end_x, end_y) = crop_percentage_to_pixel(
            img.get_width(), img.get_height(), crop
        )
        if end_x - start_x > 0 and end_y - start_y > 0:
            img = img.new_subpixbuf(start_x, start_y, end_x - start_x, end_y - start_y)

    # Need to keep the ratio intact, otherwise scaled images look stretched
    # if the dimensions aren't close in size
    (width, height) = image_actual_size(
        width, height, img.get_width(), img.get_height()
    )

    scaled = img.scale_simple(int(width), int(height), GdkPixbuf.InterpType.BILINEAR)
    scaled.savev(destination, "jpeg", "", "")


# -------------------------------------------------------------------------
#
# image_dpi
#
# -------------------------------------------------------------------------
MM_PER_INCH = 25.4


def image_dpi(source):
    """
    Return the dpi found in the image header. Use a sensible
    default of the screen DPI or 96.0 dpi if N/A.

    :param source: source image file, in any format that PIL recognizes
    :type source: unicode
    :rtype: int
    :returns: (x_dpi, y_dpi)
    """
    try:
        import PIL.Image
    except ImportError:
        import logging

        logging.warning(
            _(
                "WARNING: PIL module not loaded.  "
                "Image cropping in report files will be impaired."
            )
        )
    else:
        try:
            img = PIL.Image.open(source)
        except IOError:
            pass
        else:
            try:
                dpi = img.info["dpi"]
                return dpi
            except (AttributeError, KeyError):
                pass
    try:
        from gi.repository import Gdk

        mon = Gdk.Display.get_default().get_primary_monitor()
        mon_geom = mon.get_geometry()
        scale = mon.get_scale_factor() * MM_PER_INCH
        dpi = (
            mon_geom.width * scale / mon.get_width_mm(),
            mon_geom.height * scale / mon.get_height_mm(),
        )
    except:
        dpi = (96.0, 96.0)  # LibOO 3.6 assumes this if image contains no DPI info
        # This isn't safe even within a single platform (Windows), but we
        # can't do better if all of the above failed. See bug# 7290.
    return dpi


# -------------------------------------------------------------------------
#
# image_size
#
# -------------------------------------------------------------------------
def image_size(source):
    """
    Return the width and size of the specified image.

    :param source: source image file, in any format that gtk recongizes
    :type source: unicode
    :rtype: tuple(int, int)
    :returns: a tuple consisting of the width and height
    """
    from gi.repository import GdkPixbuf
    from gi.repository import GLib

    try:
        # For performance reasons, we'll try to get image size from imagesize.
        import imagesize

        return imagesize.get(source)
    except (ImportError, FileNotFoundError, ValueError):
        # python-imagesize is not installed, the file does not exist, or
        # the size cannot be determined by imagesize.
        # So Trying to get image size with Gdk.
        try:
            img = GdkPixbuf.Pixbuf.new_from_file(source)
            width = img.get_width()
            height = img.get_height()
        except GLib.GError:
            width = 0
            height = 0
        return (width, height)


# -------------------------------------------------------------------------
#
# image_actual_size
#
# -------------------------------------------------------------------------
def image_actual_size(x_cm, y_cm, x, y):
    """
    Calculate what the actual width & height of the image should be.

    :param x_cm: width in centimeters
    :type source: int
    :param y_cm: height in centimeters
    :type source: int
    :param x: desired width in pixels
    :type source: int
    :param y: desired height in pixels
    :type source: int
    :rtype: tuple(int, int)
    :returns: a tuple consisting of the width and height in centimeters
    """

    ratio = float(x_cm) * float(y) / (float(y_cm) * float(x))

    if ratio < 1:
        act_width = x_cm
        act_height = y_cm * ratio
    else:
        act_height = y_cm
        act_width = x_cm / ratio

    return (act_width, act_height)


# -------------------------------------------------------------------------
#
# resize_to_buffer
#
# -------------------------------------------------------------------------
def resize_to_buffer(source, size, crop=None):
    """
    Loads the image and resizes it. Instead of saving the file, the data
    is returned in a buffer.

    :param source: source image file, in any format that gtk recognizes
    :type source: unicode
    :param size: desired size of the destination image ([width, height])
    :type size: list
    :param crop: cropping coordinates
    :type crop: array of integers ([start_x, start_y, end_x, end_y])
    :rtype: buffer of data
    :returns: raw data
    """
    from gi.repository import GdkPixbuf

    img = GdkPixbuf.Pixbuf.new_from_file(source)

    if crop:
        (start_x, start_y, end_x, end_y) = crop_percentage_to_pixel(
            img.get_width(), img.get_height(), crop
        )
        if end_x - start_x > 0 and end_y - start_y > 0:
            img = img.new_subpixbuf(start_x, start_y, end_x - start_x, end_y - start_y)

    # Need to keep the ratio intact, otherwise scaled images look stretched
    # if the dimensions aren't close in size
    (size[0], size[1]) = image_actual_size(
        size[0], size[1], img.get_width(), img.get_height()
    )

    scaled = img.scale_simple(int(size[0]), int(size[1]), GdkPixbuf.InterpType.BILINEAR)

    return scaled


# -------------------------------------------------------------------------
#
# resize_to_jpeg_buffer
#
# -------------------------------------------------------------------------
def resize_to_jpeg_buffer(source, size, crop=None):
    """
    Loads the image, converting the file to JPEG, and resizing it. Instead of
    saving the file, the data is returned in a buffer.

    :param source: source image file, in any format that gtk recognizes
    :type source: unicode
    :param size: desired size of the destination image ([width, height])
    :type size: list
    :param crop: cropping coordinates
    :type crop: array of integers ([start_x, start_y, end_x, end_y])
    :rtype: buffer of data
    :returns: jpeg image as raw data
    """
    from gi.repository import GdkPixbuf

    filed, dest = tempfile.mkstemp()
    img = GdkPixbuf.Pixbuf.new_from_file(source)

    if crop:
        (start_x, start_y, end_x, end_y) = crop_percentage_to_pixel(
            img.get_width(), img.get_height(), crop
        )
        if end_x - start_x > 0 and end_y - start_y > 0:
            img = img.new_subpixbuf(start_x, start_y, end_x - start_x, end_y - start_y)

    # Need to keep the ratio intact, otherwise scaled images look stretched
    # if the dimensions aren't close in size
    (size[0], size[1]) = image_actual_size(
        size[0], size[1], img.get_width(), img.get_height()
    )

    scaled = img.scale_simple(int(size[0]), int(size[1]), GdkPixbuf.InterpType.BILINEAR)
    os.close(filed)
    scaled.savev(dest, "jpeg", "", "")
    with open(dest, mode="rb") as ofile:
        data = ofile.read()
    try:
        os.unlink(dest)
    except:
        pass
    return data



# -*- coding: utf-8 -*-
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2003-2006  Donald N. Allingham
# Copyright (C) 2008       Brian G. Matherly
# Copyright (C) 2010       Andrew I Baznikin
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

#
# Written by Egyeki Gergely <egeri@elte.hu>, 2004
"""
Specific classes for relationships.
"""

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------

from gramps.gen.lib import Person
import gramps.gen.relationship

# -------------------------------------------------------------------------
#
# Shared constants
#
# -------------------------------------------------------------------------

_level = [
    "",
    "",
    "másod",
    "harmad",
    "negyed",
    "ötöd",
    "hatod",
    "heted",
    "nyolcad",
    "kilenced",
    "tized",
    "tizenegyed",
    "tizenketted",
    "tizenharmad",
    "tizennegyed",
    "tizenötöd",
    "tizenhatod",
    "tizenheted",
    "tizennyolcad",
    "tizenkilenced",
    "huszad",
    "huszonegyed",
]


# -------------------------------------------------------------------------
#
# Specific relationship functions
#
# -------------------------------------------------------------------------
class RelationshipCalculator(gramps.gen.relationship.RelationshipCalculator):
    """
    RelationshipCalculator Class
    """

    def __init__(self):
        gramps.gen.relationship.RelationshipCalculator.__init__(self)

    def get_parents(self, level):
        if level == 0:
            return ""
        elif level == 1:
            return "szülei"
        elif level == 2:
            return "nagyszülei"
        elif level == 3:
            return "dédszülei"
        elif level == 4:
            return "ükszülei"
        else:
            return "%d. szülei" % level

    def get_father(self, level):
        if level == 0:
            return ""
        elif level == 1:
            return "apja"
        elif level == 2:
            return "nagyapja"
        elif level == 3:
            return "dédapja"
        elif level == 4:
            return "ükapja"
        else:
            return "%d. ükapja" % level

    def get_mother(self, level):
        if level == 0:
            return ""
        elif level == 1:
            return "anyja"
        elif level == 2:
            return "nagyanyja"
        elif level == 3:
            return "dédanyja"
        elif level == 4:
            return "ükanyja"
        else:
            return "%d. ükanyja" % level

    def get_son(self, level):
        if level == 0:
            return ""
        elif level == 1:
            return "fia"
        elif level == 2:
            return "unokája"
        elif level == 3:
            return "dédunokája"
        elif level == 4:
            return "ükunokája"
        else:
            return "%d. unokája" % level

    def get_daughter(self, level):
        if level == 0:
            return ""
        elif level == 1:
            return "lánya"
        elif level <= len([level]):
            return self.get_son(level)

    def get_uncle(self, level):
        if level == 0:
            return ""
        elif level == 1:
            return "testvére"
        elif level == 2:
            return "nagybátyja"
        else:
            return "%d. nagybátyja" % level

    def get_aunt(self, level):
        if level == 0:
            return ""
        elif level == 1:
            return "testvére"
        elif level == 2:
            return "nagynénje"
        else:
            return "%d. nagynénje" % level

    def get_nephew(self, level):
        if level == 0:
            return ""
        elif level == 1:
            return "unokája"
        else:
            return "%d. unokája" % level

    def get_niece(self, level):
        return self.get_nephew(level)

    def get_male_cousin(self, level):
        if level == 0:
            return ""
        elif level == 1:
            return "unokatestvére"
        else:
            return "%d. unokatestvére" % level

    def get_female_cousin(self, level):
        return self.get_male_cousin(level)

    # ----------------------------------------------
    #
    # brother and sister age differences
    #
    # ----------------------------------------------

    def get_age_comp(self, orig_person, other_person):
        # in 3.X api we can't know persons age
        return 0

    def get_age_brother(self, level):
        if level == 0:
            return "testvére"
        elif level == 1:
            return "öccse"
        else:
            return "bátyja"

    def get_age_sister(self, level):
        if level == 0:
            return "testvére"
        elif level == 1:
            return "húga"
        else:
            return "nővére"

    # ---------------------------------------------
    #
    # en: father-in-law, mother-in-law, son-in-law, daughter-in-law
    # hu: após, anyós, vő, meny
    #
    # ---------------------------------------------

    # FIXME: is it still used?
    def is_fathermother_in_law(self, db, orig, other):
        for f in other.get_family_handle_list():
            family = db.get_family_from_handle(f)
            sp_id = None
            if family:
                if other == family.get_father_handle():
                    sp_id = family.get_mother_handle()
                elif other == family.get_mother_handle():
                    sp_id = family.get_father_handle()
                for g in orig.get_family_handle_list():
                    family = db.get_family_from_handle(g)
                    if family:
                        if sp_id in family.get_child_handle_list():
                            return 1
        return 0

    # ------------------------------------------------------------------------
    #
    # hu: sógor, sógornő
    # en: brother-in-law, sister-in-law
    #
    # ------------------------------------------------------------------------

    # FIXME: is it still used?
    def is_brothersister_in_law(self, db, orig, other):
        for f in orig.get_family_handle_list():
            family = db.get_family_from_handle(f)
            sp_id = None
            if family:
                if orig == family.get_father_handle():
                    sp_id = family.get_mother_handle()
                elif other == family.get_mother_handle():
                    sp_id = family.get_father_handler()

                p = other.get_main_parents_family_handle()
                family = db.get_family_from_handle(p)
                if family:
                    c = family.get_child_handle_list()
                    if (other.get_handle() in c) and (sp_id in c):
                        return 1
        return 0

    # -------------------------------------------------------------------------
    #
    # get_relationship
    #
    # -------------------------------------------------------------------------

    def get_relationship(
        self, secondRel, firstRel, orig_person, other_person, in_law_a, in_law_b
    ):
        """
        returns a string representing the relationshp between the two people,
        along with a list of common ancestors (typically father,mother)
        """

        common = ""

        if in_law_a or in_law_a:
            if firstRel == 0 and secondRel == 0:
                if other_person == Person.MALE:
                    return ("apósa", "")
                elif other_person == Person.FEMALE:
                    return ("anyósa", "")
                else:
                    return ("apósa vagy anyósa", "")

            elif secondRel == 0:
                if orig_person == Person.MALE:
                    return ("veje", "")
                elif orig_person == Person.FEMALE:
                    return ("menye", "")
                else:
                    return ("veje vagy menye", "")

            elif firstRel == 1:
                if other_person == Person.MALE:
                    return ("sógora", "")
                elif other_person == Person.FEMALE:
                    return ("sógornője", "")
                else:
                    return ("sógora vagy sógornője", "")

        if firstRel == 0:
            if secondRel == 0:
                return ("", common)
            elif other_person == Person.MALE:
                return (self.get_father(secondRel), common)
            else:
                return (self.get_mother(secondRel), common)

        elif secondRel == 0:
            if other_person == Person.MALE:
                return (self.get_son(firstRel), common)
            else:
                return (self.get_daughter(firstRel), common)

        elif firstRel == 1:
            if other_person == Person.MALE:
                if secondRel == 1:
                    return (
                        self.get_age_brother(
                            self.get_age_comp(orig_person, other_person)
                        ),
                        common,
                    )
                else:
                    return (self.get_uncle(secondRel), common)
            else:
                if secondRel == 1:
                    return (
                        self.get_age_sister(
                            self.get_age_comp(orig_person, other_person)
                        ),
                        common,
                    )
                else:
                    return (self.get_aunt(secondRel), common)

        elif secondRel == 1:
            if other_person == Person.MALE:
                return (self.get_nephew(firstRel - 1), common)
            else:
                return (self.get_niece(firstRel - 1), common)

        else:
            if other_person == Person.MALE:
                return (self.get_male_cousin(firstRel - 1), common)
            else:
                return (self.get_female_cousin(firstRel - 1), common)

    def get_single_relationship_string(
        self,
        Ga,
        Gb,
        gender_a,
        gender_b,
        reltocommon_a,
        reltocommon_b,
        only_birth=True,
        in_law_a=False,
        in_law_b=False,
    ):
        return self.get_relationship(Ga, Gb, gender_a, gender_b, in_law_a, in_law_b)[0]

    def get_sibling_relationship_string(
        self, sib_type, gender_a, gender_b, in_law_a=False, in_law_b=False
    ):
        return self.get_relationship(1, 1, gender_a, gender_b, in_law_a, in_law_b)[0]


if __name__ == "__main__":
    # Test function. Call it as follows from the command line (so as to find
    #        imported modules):
    #    export PYTHONPATH=/path/to/gramps/src
    # python src/plugins/rel/rel_hu.py
    # (Above not needed here)

    """TRANSLATORS, copy this if statement at the bottom of your
    rel_xx.py module, and test your work with:
    python src/plugins/rel/rel_xx.py
    """
    from gramps.gen.relationship import test

    RC = RelationshipCalculator()
    test(RC, True)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2015        Nick Hall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

__all__ = ["PlaceEntry"]

# -------------------------------------------------------------------------
#
# Standard python modules
#
# -------------------------------------------------------------------------
import logging

_LOG = logging.getLogger(".widgets.placeentry")

# -------------------------------------------------------------------------
#
# GTK/Gnome modules
#
# -------------------------------------------------------------------------
from gi.repository import Gtk

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from ..selectors import SelectorFactory


# -------------------------------------------------------------------------
#
# PlaceEntry class
#
# -------------------------------------------------------------------------
class PlaceEntry(Gtk.Box):
    def __init__(self, dbstate, uistate, track):
        Gtk.Box.__init__(self)
        self.dbstate = dbstate
        self.uistate = uistate
        self.track = track
        self.entry = Gtk.Entry()
        self.entry.set_width_chars(5)
        self.pack_start(self.entry, True, True, 0)
        image = Gtk.Image()
        image.set_from_icon_name("gtk-index", Gtk.IconSize.BUTTON)
        button = Gtk.Button()
        button.set_image(image)
        button.set_relief(Gtk.ReliefStyle.NORMAL)
        self.pack_start(button, False, True, 0)
        button.connect("clicked", self.on_clicked)
        self.show_all()

    def on_clicked(self, button):
        SelectPlace = SelectorFactory("Place")
        sel = SelectPlace(self.dbstate, self.uistate, self.track)
        place = sel.run()
        if place:
            self.set_text(place.gramps_id)

    def get_text(self):
        return self.entry.get_text()

    def set_text(self, text):
        self.entry.set_text(text)



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2009       Benny Malengier
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
"""
OpenStreetMap map service plugin. Open place in openstreetmap.org
"""

# ------------------------------------------------------------------------
#
# python modules
#
# ------------------------------------------------------------------------

# ------------------------------------------------------------------------
#
# Gramps modules
#
# ------------------------------------------------------------------------
from gramps.plugins.lib.libmapservice import MapService
from gramps.gen.utils.location import get_main_location
from gramps.gen.display.place import displayer as place_displayer
from gramps.gen.lib import PlaceType


class OpensStreetMapService(MapService):
    """Map  service using http://openstreetmap.org
    Resource: http://wiki.openstreetmap.org/wiki/Nominatim
    """

    def __init__(self):
        MapService.__init__(self)

    def calc_url(self):
        """Determine the url to use
        Logic: Use lat lon if present
               otherwise use description of the place
        """
        place = self._get_first_place()[0]
        latitude, longitude = self._lat_lon(place)
        if longitude and latitude:
            self.url = "http://www.openstreetmap.org/" "?lat=%s&lon=%s&zoom=15" % (
                latitude,
                longitude,
            )

            return

        titledescr = place_displayer.display(self.database, place)
        self.url = "http://nominatim.openstreetmap.org/" "search?q=%s" % "+".join(
            titledescr.split()
        )



# -*- coding: utf-8 -*-
#!/usr/bin/env python
#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2007  Donald N. Allingham
# Copyright (C) 2007       Johan Gonqvist <johan.gronqvist@gmail.com>
# Copyright (C) 2007-2009  Gary Burton <gary.burton@zen.co.uk>
# Copyright (C) 2007-2009  Stephane Charette <stephanecharette@gmail.com>
# Copyright (C) 2008-2009  Brian G. Matherly
# Copyright (C) 2008       Jason M. Simanek <jason@bohemianalps.com>
# Copyright (C) 2008-2011  Rob G. Healey <robhealey1@gmail.com>
# Copyright (C) 2010       Doug Blank <doug.blank@gmail.com>
# Copyright (C) 2010       Jakim Friant
# Copyright (C) 2010-      Serge Noiraud
# Copyright (C) 2011       Tim G L Lyons
# Copyright (C) 2013       Benny Malengier
# Copyright (C) 2016       Allen Crider
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

"""
Narrative Web Page generator.

Classe:
    CitationPages - dummy
"""
# ------------------------------------------------
# python modules
# ------------------------------------------------
from decimal import getcontext
import logging

# ------------------------------------------------
# Gramps module
# ------------------------------------------------

# ------------------------------------------------
# specific narrative web import
# ------------------------------------------------
from gramps.plugins.webreport.basepage import BasePage

LOG = logging.getLogger(".NarrativeWeb")
getcontext().prec = 8


#################################################
#
#    Passes citations through to the Sources page
#
#################################################
class CitationPages(BasePage):
    """
    This class is responsible for displaying information about the 'Citation'
    database objects. It passes this information to the 'Sources' tab. It is
    told by the 'add_instances' call which 'Citation's to display.
    """

    def __init__(self, report, the_lang, the_title):
        """
        @param: report    -- The instance of the main report class
                             for this report
        @param: the_lang  -- The lang to process
        @param: the_title -- The title page related to the language
        """
        BasePage.__init__(self, report, the_lang, the_title)

    def display_pages(self, the_lang, the_title):
        pass



#
# Gramps - a GTK+/GNOME based genealogy program
#
# Copyright (C) 2000-2006  Donald N. Allingham
# Copyright (C) 2009       Gary Burton
# Copyright (C) 2009       Benny Malengier
# Copyright (C) 2010       Nick Hall
# Copyright (C) 2011       Tim G L Lyons
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# -------------------------------------------------------------------------
#
# Python classes
#
# -------------------------------------------------------------------------
import logging

_LOG = logging.getLogger(".gui.editors.EditNote")

# -------------------------------------------------------------------------
#
# GTK libraries
#
# -------------------------------------------------------------------------
from gi.repository import Gtk
from gi.repository import GObject
from gi.repository import Pango

# -------------------------------------------------------------------------
#
# Gramps modules
#
# -------------------------------------------------------------------------
from gramps.gen.const import GRAMPS_LOCALE as glocale

_ = glocale.translation.sgettext
from gramps.gen.config import config
from .editprimary import EditPrimary
from .displaytabs import GrampsTab, NoteBackRefList
from ..widgets import (
    MonitoredDataType,
    MonitoredCheckbox,
    MonitoredEntry,
    PrivacyButton,
    MonitoredTagList,
)
from gramps.gen.lib import Note
from gramps.gen.db import DbTxn
from ..dialog import ErrorDialog
from ..glade import Glade
from gramps.gen.const import URL_MANUAL_SECT2

# -------------------------------------------------------------------------
#
# Constants
#
# -------------------------------------------------------------------------

WIKI_HELP_PAGE = URL_MANUAL_SECT2
WIKI_HELP_SEC = _("Editing_information_about_notes", "manual")

# -------------------------------------------------------------------------
#
# NoteTab
#
# -------------------------------------------------------------------------


class NoteTab(GrampsTab):
    """
    This class provides the tabpage of the note
    """

    def __init__(self, dbstate, uistate, track, name, widget):
        """
        @param dbstate: The database state. Contains a reference to
        the database, along with other state information. The GrampsTab
        uses this to access the database and to pass to and created
        child windows (such as edit dialogs).
        @type dbstate: L{DbState.DbState}
        @param uistate: The UI state. Used primarily to pass to any created
        subwindows.
        @type uistate: L{DisplayState.DisplayState}
        @param track: The window tracking mechanism used to manage windows.
        This is only used to pass to generted child windows.
        @type track: list
        @param name: Notebook label name
        @type name: str/unicode
        @param widget: widget to be shown in the tab
        @type widget: gtk widget
        """
        GrampsTab.__init__(self, dbstate, uistate, track, name)
        eventbox = Gtk.EventBox()
        eventbox.add(widget)
        self.pack_start(eventbox, True, True, 0)
        self._set_label(show_image=False)
        eventbox.connect("key_press_event", self.key_pressed)
        self.show_all()

    def is_empty(self):
        """
        Override base class
        """
        return False


# -------------------------------------------------------------------------
#
# EditNote
#
# -------------------------------------------------------------------------


class EditNote(EditPrimary):
    def __init__(
        self,
        dbstate,
        uistate,
        track,
        note,
        callback=None,
        callertitle=None,
        extratype=None,
    ):
        """Create an EditNote window. Associate a note with the window.

        @param callertitle: Text passed by calling object to add to title
        @type callertitle: str
        @param extratype: Extra L{NoteType} values to add to the default types.
        They are removed from the ignorelist of L{NoteType}.
        @type extratype: list of int

        """
        self.callertitle = callertitle
        self.extratype = extratype
        EditPrimary.__init__(
            self,
            dbstate,
            uistate,
            track,
            note,
            dbstate.db.get_note_from_handle,
            dbstate.db.get_note_from_gramps_id,
            callback,
        )

    def empty_object(self):
        """Return an empty Note object for comparison for changes.

        It is used by the base class L{EditPrimary}.

        """
        empty_note = Note()
        if self.extratype:
            empty_note.set_type(self.extratype[0])
        return empty_note

    def get_menu_title(self):
        if self.obj.get_handle():
            if self.callertitle:
                title = _("Note: %(id)s - %(context)s") % {
                    "id": self.obj.get_gramps_id(),
                    "context": self.callertitle,
                }
            else:
                title = _("Note: %s") % self.obj.get_gramps_id()
        else:
            if self.callertitle:
                title = _("New Note - %(context)s") % {"context": self.callertitle}
            else:
                title = _("New Note")

        return title

    def get_custom_notetypes(self):
        return self.dbstate.db.get_note_types()

    def _local_init(self):
        """Local initialization function.

        Perform basic initialization, including setting up widgets
        and the glade interface. It is called by the base class L{EditPrimary},
        and overridden here.

        """
        self.top = Glade()

        win = self.top.toplevel
        self.set_window(win, None, self.get_menu_title())
        self.setup_configs("interface.note", 700, 500)

        vboxnote = self.top.get_object("vbox131")
        notebook = self.top.get_object("note_notebook")
        # recreate start page as GrampsTab
        notebook.remove_page(0)
        self.ntab = NoteTab(
            self.dbstate, self.uistate, self.track, _("_Note"), vboxnote
        )
        self.track_ref_for_deletion("ntab")

        self.build_interface()

    def _setup_fields(self):
        """Get control widgets and attach them to Note's attributes."""
        self.type_selector = MonitoredDataType(
            self.top.get_object("type"),
            self.obj.set_type,
            self.obj.get_type,
            self.db.readonly,
            custom_values=self.get_custom_notetypes(),
            ignore_values=self.obj.get_type().get_ignore_list(self.extratype),
        )

        self.check = MonitoredCheckbox(
            self.obj,
            self.top.get_object("format"),
            self.obj.set_format,
            self.obj.get_format,
            readonly=self.db.readonly,
        )

        self.gid = MonitoredEntry(
            self.top.get_object("id"),
            self.obj.set_gramps_id,
            self.obj.get_gramps_id,
            self.db.readonly,
        )

        self.tags = MonitoredTagList(
            self.top.get_object("tag_label"),
            self.top.get_object("tag_button"),
            self.obj.set_tag_list,
            self.obj.get_tag_list,
            self.db,
            self.uistate,
            self.track,
            self.db.readonly,
        )

        self.priv = PrivacyButton(
            self.top.get_object("private"), self.obj, self.db.readonly
        )

    def _connect_signals(self):
        """Connects any signals that need to be connected.

        Called by the init routine of the base class L{EditPrimary}.

        """
        self.define_ok_button(self.top.get_object("ok"), self.save)
        self.define_cancel_button(self.top.get_object("cancel"))
        self.define_help_button(
            self.top.get_object("help"), WIKI_HELP_PAGE, WIKI_HELP_SEC
        )

    def _connect_db_signals(self):
        """
        Connect any signals that need to be connected.
        Called by the init routine of the base class (_EditPrimary).
        """
        self._add_db_signal("note-rebuild", self._do_close)
        self._add_db_signal("note-delete", self.check_for_close)

    def _create_tabbed_pages(self):
        """Create the notebook tabs and inserts them into the main window."""
        notebook = self.top.get_object("note_notebook")

        self._add_tab(notebook, self.ntab)

        handles = self.dbstate.db.find_backlink_handles(self.obj.handle)
        self.rlist = NoteBackRefList(
            self.dbstate, self.uistate, self.track, handles, "note_editor_references"
        )
        self.backref_tab = self._add_tab(notebook, self.rlist)
        self.track_ref_for_deletion("rlist")
        self.track_ref_for_deletion("backref_tab")

        self._setup_notebook_tabs(notebook)

    def build_interface(self):
        self.texteditor = self.top.get_object("texteditor")
        self.texteditor.set_editable(not self.dbstate.db.readonly)
        self.texteditor.set_wrap_mode(Gtk.WrapMode.WORD)

        # create a formatting toolbar
        if not self.dbstate.db.readonly:
            vbox = self.top.get_object("container")
            toolbar, self.action_group = self.texteditor.create_toolbar(
                self.uistate.uimanager, self.window
            )
            vbox.pack_start(toolbar, False, False, 0)
            self.texteditor.set_transient_parent(self.window)

        # setup initial values for textview and textbuffer
        if self.obj:
            self.empty = False
            with self.texteditor.undo_disabled():
                self.texteditor.set_text(self.obj.get_styledtext())
            # Reset the undoable buffer:
            self.texteditor.reset()
            _LOG.debug("Initial Note: %s" % str(self.texteditor.get_text()))
        else:
            self.empty = True

    def build_menu_names(self, person):
        """
        Provide the information needed by the base class to define the
        window management menu entries.
        """
        return (_("Edit Note"), self.get_menu_title())

    def _post_init(self):
        self.texteditor.grab_focus()

    def update_note(self):
        """Update the Note object with current value."""
        if self.obj:
            text = self.texteditor.get_text()
            self.obj.set_styledtext(text)
            _LOG.debug(str(text))

    def close(self, *obj):
        """Called when cancel button clicked."""
        self.update_note()
        super().close()

    def save(self, *obj):
        """Save the data."""
        self.ok_button.set_sensitive(False)

        self.update_note()

        if self.object_is_empty():
            ErrorDialog(
                _("Cannot save note"),
                _(
                    "No data exists for this note. Please "
                    "enter data or cancel the edit."
                ),
                parent=self.window,
            )
            self.ok_button.set_sensitive(True)
            return

        (uses_dupe_id, id) = self._uses_duplicate_id()
        if uses_dupe_id:
            msg1 = _("Cannot save note. ID already exists.")
            msg2 = _(
                "You have attempted to use the existing Gramps ID with "
                "value %(id)s. This value is already used. Please "
                "enter a different ID or leave "
                "blank to get the next available ID value."
            ) % {"id": id}
            ErrorDialog(msg1, msg2, parent=self.window)
            self.ok_button.set_sensitive(True)
            return

        if not self.obj.handle:
            with DbTxn(_("Add Note"), self.db) as trans:
                self.db.add_note(self.obj, trans)
        else:
            if self.data_has_changed():
                with DbTxn(_("Edit Note"), self.db) as trans:
                    if not self.obj.get_gramps_id():
                        self.obj.set_gramps_id(self.db.find_next_note_gramps_id())
                    self.db.commit_note(self.obj, trans)

        self._do_close()
        if self.callback:
            self.callback(self.obj.get_handle())
